using System;
using System.IO;			// for StreamWriter I/O to a file
using System.Windows.Forms;	// for RichTextBox

namespace MyNamespace
{
	/// <summary>
	///		Extra precision floating-point List algorithms
	///		Developed in TURBO Pascal 5.5 and 6.0, converted to Visual Basic 6
	///		and now in Visual C# .Net
	///
	/// ---------------- *** LEGAL NOTICE: *** ------------------
	///
	/// This program is free software; you can redistribute it and/or modify
	/// it under the terms of the GNU General Public License (GPL) as published
	/// by the Free Software Foundation. cf. the file COPYING.txt.
	///
	/// ------------ *** end of legal notice *** ------------------
	/// </summary>

	public class XMCalLst
	{
		public XMCalLst()
		{
			uFirst = 1;
		}

		public static void Trace(MultiM x, MultiM z) // z = sum of elements on the principal diagonal
		{
			if (x.Rs == 0)
				MultiMD.ErrNotMatrix();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiF sum = new MultiF(Calc.fMC);
				int n = Math.Min(x.Rs, x.Cs);
				for (int i = 0; i < n; i++)
					sum += x.E[i, i];
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.SetToNM(sum, z);
			}
		} // Trace

		public static void TraceN(MultiM x, int zI) // z = sum of elements on the principal diagonal
		{
			if (x.Rs == 0)
				MultiMD.ErrNotMatrix();
			else
			{
				MultiM z = new MultiM();
				Trace(x, z);
				Calc.SetToN(z, zI);			}
		} // TraceN

		public static void IsDiagN(MultiM x, int zI) // is x a diagonal matrix 1 (True), 0 (False
		{
			if (x.Rs == 0 || x.Rs != x.Cs)
				Calc.ClearN(zI);
			else
			{
				int i = 1;
				for (int r = 0; r < x.Rs; r++)
				{
					for (int c = 0; c < x.Cs; c++)
					{
						if (r != c && x.E[r, c] != MultiFD.fZero)
						{
							i = 0;
							break;
						}
					}
					if (i == 0)
						break;
				}
				Calc.ClearN(zI);
				Common.nList[zI].V[0] = i;
			}
		} // IsDiagN

		public static void MatIdN(MultiM x, int zI) // z = x by x identity matrix
		{
			int n;
			MultiFD.Get1F(out n, x);
			if (x.Rs != 0)
				MultiMD.ErrNotScalar();
			else if (n < 1)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "MatIdN-tI", out tI);
				MatrixN(n, n, tI);
				for (int r = 0; r < n; r++)
					for (int c = 0; c < n; c++)
						Common.nList[tI].E[r, c] &= MultiFD.fZero;
				for (int i = 0; i < n; i++)
					Common.nList[tI].E[i, i] &= MultiFD.fOne;
				Calc.SetToII(tI, zI);
				Common.LDelete(tI);
			}
		} // MatIdN

		public static void HilbertN(MultiM x, int zI) // z = x by x Hilbert matrix
		// zij = 1/(i+j-1)
		{
			int n;
			MultiFD.Get1F(out n, x);
			if (x.Rs != 0)
				MultiMD.ErrNotScalar();
			else if (n < 1)
				MultiMD.ErrIncompatible();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiM z = new MultiM(Calc.fMC, n, n);
				for (int i = 0; i < n; i++)
					for (int j = 0; j < n; j++)
						MultiFD.Div1F(MultiFD.fOne, i + j + 1, z.E[i, j]);
				Calc.SetToN(z, zI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
			}
		} // HilbertN

		public static void PascalN(MultiM x, int zI) // z = x by x Pascal triangle
		// zij = Bino(i, j)
		{
			int n;
			MultiFD.Get1F(out n, x);
			if (x.Rs != 0)
				MultiMD.ErrNotScalar();
			else if (n < 1)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "PascalN-tI", out tI);
				MultiF iF = new MultiF(1);
				MultiF jF = new MultiF(1);
				MatrixN(n, n, zI);
				for (int i = 0; i < n; i++)
				{
					iF &= i;
					for (int j = 0; j < n; j++)
					{
						jF &= j;
						XPCalLst.BinoN(iF, jF, tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM(Common.nList[tI], Common.nList[zI].E[i, j]);
					}
				}
				Common.LDelete(tI);
			}
		} // PascalN

		public static void DiagonalN(MultiM x, int zI) // z = diagonal matrix from vector x
		{
			Calc.SetToN(MultiMD.Diagonal(x), zI);
		} // DiagonalN

		public static void Del(MultiM x, int c, MultiM z) // delete column c from matrix x
		{
			if (c < 0 || c >= x.Cs)
				Calc.SetToNM(x, z);
			else if (x.Cs == 1)
				Calc.SetToNM(MultiFD.fZero, z);
			else
			{
				MultiM zz = new MultiM(1, x.Rs, x.Cs - 1);
				for (int r = 0; r < x.Rs; r++)
					for (int ci = 0; ci < c; ci++)
						XPCalLst.SetToNM(x.E[r, ci], zz.E[r, ci]);
				for (int r = 0; r < x.Rs; r++)
					for (int ci = c + 1; ci < x.Cs; ci++)
						XPCalLst.SetToNM(x.E[r, ci], zz.E[r, ci - 1]);
				Calc.SetToNM(zz, z);
			}
		} // Del

		public static void DelN(MultiM x, MultiM y, int zI) // delete column y from matrix x
		{
			int c;
			MultiFD.Get1F(out c, y);
			if (x.Rs == 0)
				MultiMD.ErrNotMatrix();
			else if (y.Rs != 0)
				MultiMD.ErrNotScalar();
			else if (c < 1 || c > x.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				c -= uFirst;
				MultiM zz = new MultiM();
				Del(x, c, zz);
				Calc.SetToN(zz, zI);
			}
		} // DelN

		public static void DelR(MultiM x, int r, MultiM z) // delete row r from matrix x
		{
			if (r < 0 || r >= x.Rs)
				Calc.SetToNM(x, z);
			else if (x.Rs == 1)
				Calc.SetToNM(MultiFD.fZero, z);
			else
			{
				MultiM zz = new MultiM(1, x.Rs - 1, x.Cs);
				for (int ri = 0; ri < r; ri++)
					for (int c = 0; c < x.Cs; c++)
						XPCalLst.SetToNM(x.E[ri, c], zz.E[ri, c]);
				for (int ri = r + 1; ri < x.Rs; ri++)
					for (int c = 0; c < x.Cs; c++)
						XPCalLst.SetToNM(x.E[ri, c], zz.E[ri - 1, c]);
				Calc.SetToNM(zz, z);
			}
		} // DelR

		public static void DelRN(MultiM x, MultiM y, int zI) // delete row y from matrix x
		{
			int r;
			MultiFD.Get1F(out r, y);
			if (x.Rs == 0)
				MultiMD.ErrNotMatrix();
			else if (y.Rs != 0)
				MultiMD.ErrNotScalar();
			else if (r < 1 || r > x.Rs)
				MultiMD.ErrIncompatible();
			else
			{
				r -= uFirst;
				MultiM zz = new MultiM();
				DelR(x, r, zz);
				Calc.SetToN(zz, zI);
			}
		} // DelRN

		public static void ExtractN(MultiM x, MultiM y, int zI) // extract column y from matrix x
		{
			int c;
			MultiFD.Get1F(out c, y);
			if (x.Rs == 0)
				MultiMD.ErrNotMatrix();
			else if (y.Rs != 0)
				MultiMD.ErrNotScalar();
			else if (c < 1 || c > x.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				c -= uFirst;
				int tI;
				Common.InitNAdd(1, "ExtractN-tI", out tI);
				MatrixN(x.Rs, 1, tI);
				for (int r = 0; r < x.Rs; r++)
					XPCalLst.SetToNM(x.E[r, c], Common.nList[tI].E[r, 0]);
				Calc.SetToII(tI, zI);
				Common.LDelete(tI);
			}
		} // ExtractN

		public static void ExtractRN(MultiM x, MultiM y, int zI) // extract row y from matrix x
		{
			int r;
			MultiFD.Get1F(out r, y);
			if (x.Rs == 0)
				MultiMD.ErrNotMatrix();
			else if (y.Rs != 0)
				MultiMD.ErrNotScalar();
			else if (r < 1 || r > x.Rs)
				MultiMD.ErrIncompatible();
			else
			{
				r -= uFirst;
				int tI;
				Common.InitNAdd(1, "ExtractRN-tI", out tI);
				MatrixN(1, x.Cs, tI);
				for (int c = 0; c < x.Cs; c++)
					XPCalLst.SetToNM(x.E[r, c], Common.nList[tI].E[0, c]);
				Calc.SetToII(tI, zI);
				Common.LDelete(tI);
			}
		} // ExtractRN

		public static void ConcatN(MultiM x, MultiM y, int zI) // z = [x y]
		{
			if (x.Rs == 0 || y.Rs == 0)
				MultiMD.ErrNotMatrix();
			else if (x.Rs != y.Rs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "ConcatN-tI", out tI);
				MatrixN(x.Rs, x.Cs + y.Cs, tI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
						XPCalLst.SetToNM(x.E[r, c], Common.nList[tI].E[r, c]);
				for (int r = 0; r < y.Rs; r++)
					for (int c = 0; c < y.Cs; c++)
						XPCalLst.SetToNM(y.E[r, c], Common.nList[tI].E[r, c+x.Cs]);
				Calc.SetToII(tI, zI);
				Common.LDelete(tI);
			}
		} // ConcatN

		public static void ConcatRN(MultiM x, MultiM y, int zI) // z =	[x]
		{														//		[y]
			if (x.Rs == 0 || y.Rs == 0)
				MultiMD.ErrNotMatrix();
			else if (x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "ConcatRN-tI", out tI);
				MatrixN(x.Rs + y.Rs, x.Cs, tI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
						XPCalLst.SetToNM(x.E[r, c], Common.nList[tI].E[r, c]);
				for (int r = 0; r < y.Rs; r++)
					for (int c = 0; c < y.Cs; c++)
						XPCalLst.SetToNM(y.E[r, c], Common.nList[tI].E[r + x.Rs, c]);
				Calc.SetToII(tI, zI);
				Common.LDelete(tI);
			}
		} // ConcatRN

		public static void ContFracN(MultiM x, MultiM y, int zI)
		{ // z = x as continuous fraction row vector [b0, b1, b2,..., by-1]
			if (x.Rs != 0 || y.Rs != 0)
				MultiMD.ErrNotScalar();
			else
			{
				MultiF xL = new MultiF(Calc.fMC);
				xL &= x;
				int n;
				MultiFD.Get1F(out n, y);
				if (n < 1)
					n = Int16.MaxValue;
				int tI;
				Common.InitNAdd(1, "ContFracN-tI", out tI);
				MatrixN(1, n, tI);
				MultiF temp = new MultiF(Calc.fMC);
				MultiFD.FloorF(xL, temp);
				XPCalLst.SetToNM(temp, Common.nList[tI].E[0, 0]);
				MultiF z2 = new MultiF(Calc.fMC);
				MultiF z1 = new MultiF(Calc.fMC);
				MultiF aM2 = new MultiF(Calc.fMC);
				MultiF aM1 = new MultiF(Calc.fMC);
				MultiF bM2 = new MultiF(Calc.fMC);
				MultiF bM1 = new MultiF(Calc.fMC);
				MultiF aK = new MultiF(Calc.fMC);
				MultiF bK = new MultiF(Calc.fMC);
				int c;
				for (c = 1; c < n; c++)
				{
					if (c == 1)
					{
						aM1 &= MultiFD.fOne;
						bM1 &= MultiFD.fZero;
						aK &= temp;
						bK &= MultiFD.fOne;
					}
					xL -= temp;
					if (xL == MultiFD.fZero)
						break;
					MultiFD.DivF(MultiFD.fOne, xL, xL); // x = 1 / x
					MultiFD.FloorF(xL, temp);
					if (temp.M - temp.N <= 2)
						break;
					XPCalLst.SetToNM(temp, Common.nList[tI].E[0, c]);
					aM2 &= aM1;
					bM2 &= bM1;
					aM1 &= aK;
					bM1 &= bK;
					aK &= temp * aM1 + aM2;
					bK &= temp * bM1 + bM2;
					z2 &= aK / bK;
					if (z1 == z2)
						break;
					z1 &= z2;
				}
				if (c < n)
				{
					Common.nList[tI].Cs = c;
				}
				n = Common.nList[tI].Cs;
				if (n > 1 && Common.nList[tI].E[0, n - 1] == MultiFD.fOne)
				{
					Common.nList[tI].E[0, n - 2] += MultiFD.fOne;
					Common.nList[tI].Cs--;
				}
				Calc.SetToII(tI, zI);
				Common.LDelete(tI);
			}
		} // ContFracN

		public static void ContFracVN(MultiM x, int zI)
		// z = Value of continuous fraction x
		{
			if (x.Rs != 1 && x.Cs != 1)
				MultiMD.ErrNotVector();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int n = Math.Max(x.Rs, x.Cs);
				MultiF aM2 = new MultiF(Calc.fMC);
				MultiF aM1 = new MultiF(Calc.fMC);
				MultiF bM2 = new MultiF(Calc.fMC);
				MultiF bM1 = new MultiF(Calc.fMC);
				MultiF aK = new MultiF(Calc.fMC);
				MultiF bK = new MultiF(Calc.fMC);
				aM1 &= MultiFD.fOne;
				bM1 &= MultiFD.fZero;
				aK &= x.E[0, 0];
				bK &= MultiFD.fOne;
				for (int i = 1; i < n; i++)
				{
					aM2 &= aM1;
					bM2 &= bM1;
					aM1 &= aK;
					bM1 &= bK;
					if (x.Rs == 1)
					{
						aK &= x.E[0, i] * aM1 + aM2;
						bK &= x.E[0, i] * bM1 + bM2;
					}
					else
					{
						aK &= x.E[i, 0] * aM1 + aM2;
						bK &= x.E[i, 0] * bM1 + bM2;
					}
				}
				aM1 &= aK / bK;
				Calc.SetToN(aM1, zI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
			}
		} // ContFracVN

		public static void InsertN(MultiM x, MultiM y, MultiM cF, int zI)
		// z = x with y inserted at column c
		{
			int c;
			MultiFD.Get1F(out c, cF);
			if (x.Rs == 0 || y.Rs == 0)
				MultiMD.ErrNotMatrix();
			else if (cF.Rs != 0)
				MultiMD.ErrNotScalar();
			else if (x.Rs != y.Rs)
				MultiMD.ErrIncompatible();
			else if (c < 1 || c > x.Cs + 1)
				MultiMD.ErrIncompatible();
			else
			{
				c -= uFirst;
				int tI;
				Common.InitNAdd(1, "InsertN-tI", out tI);
				MatrixN(x.Rs, x.Cs + y.Cs, tI);
				for (int r = 0; r < x.Rs; r++)
					for (int ci = 0; ci < c; ci++)
						XPCalLst.SetToNM(x.E[r, ci], Common.nList[tI].E[r, ci]);
				for (int r = 0; r < y.Rs; r++)
					for (int ci = 0; ci < y.Cs; ci++)
						XPCalLst.SetToNM(y.E[r, ci], Common.nList[tI].E[r, ci + c]);
				for (int r = 0; r < y.Rs; r++)
					for (int ci = c; ci < x.Cs; ci++)
						XPCalLst.SetToNM(x.E[r, ci], Common.nList[tI].E[r, ci + y.Cs]);
				Calc.SetToII(tI, zI);
				Common.LDelete(tI);
			}
		} // InsertN

		public static void InsertRN(MultiM x, MultiM y, MultiM rF, int zI)
		// z = x with y inserted at row r
		{
			int r;
			MultiFD.Get1F(out r, rF);
			if (x.Rs == 0 || y.Rs == 0)
				MultiMD.ErrNotMatrix();
			else if (rF.Rs != 0)
				MultiMD.ErrNotScalar();
			else if (x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else if (r < 1 || r > x.Rs + 1)
				MultiMD.ErrIncompatible();
			else
			{
				r -= uFirst;
				int tI;
				Common.InitNAdd(1, "InsertRN-tI", out tI);
				MatrixN(x.Rs + y.Rs, x.Cs, tI);
				for (int ri = 0; ri < r; ri++)
					for (int c = 0; c < x.Cs; c++)
						XPCalLst.SetToNM(x.E[ri, c], Common.nList[tI].E[ri, c]);
				for (int ri = 0; ri < y.Rs; ri++)
					for (int c = 0; c < y.Cs; c++)
						XPCalLst.SetToNM(y.E[ri, c], Common.nList[tI].E[ri + r, c]);
				for (int ri = r; ri < x.Rs; ri++)
					for (int c = 0; c < y.Cs; c++)
						XPCalLst.SetToNM(x.E[ri, c], Common.nList[tI].E[ri + y.Rs, c]);
				Calc.SetToII(tI, zI);
				Common.LDelete(tI);
			}
		} // InsertRN

		public static void RSizeN(int zI) // z = size of z
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				Calc.ClearN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				MatrixN(1, 2, zI);
				Common.nList[zI].E[0, 0] &= rs;
				Common.nList[zI].E[0, 1] &= cs;
			}
		}

		public static void RRIntN(int yI, int zI) // z = Random Integer between y and z inclusive
		{
			int yrs = Common.nList[yI].Rs;
			int zrs = Common.nList[zI].Rs;
			int ycs = Common.nList[yI].Cs;
			int zcs = Common.nList[zI].Cs;
			if (yrs == 0 && zrs == 0)
				XPCalLst.RRIntN(yI, zI);
			else if (yrs != zrs || ycs != zcs)
				MultiMD.ErrIncompatible();
			else
			{
				int tyI;
				int tzI;
				Common.InitNAdd(1, "RRIntN-tyI", out tyI);
				Common.InitNAdd(1, "RRIntN-tzI", out tzI);
				for (int r = 0; r < yrs; r++)
					for (int c = 0; c < ycs; c++)
					{
						XPCalLst.SetToNM(Common.nList[yI].E[r, c], (MultiF)Common.nList[tyI]);
						XPCalLst.SetToNM(Common.nList[zI].E[r, c], (MultiF)Common.nList[tzI]);
						XPCalLst.RRIntN(tyI, tzI);
						if (MultiID.muErr)
						{
							Common.LDelete(tzI);
							Common.LDelete(tyI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tzI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tzI);
				Common.LDelete(tyI);
			}		
		} // RRIntN

		public static void RAddN(MultiM x, int zI)	// z = z + x
		// sets muErr true if overflow
		{
			AddN(Common.nList[zI], x, zI);
		}

		public static void RSubN(MultiM x, int zI) // z = z - x
		// sets muErr true if overflow
		{
			SubN(Common.nList[zI], x, zI);
		}

		public static void AddN(MultiM x, MultiM y, int zI) // z = x + y
		// sets muErr true if overflow
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.AddN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
						XPCalLst.SetToNM(x.E[r, c] + y.E[r, c], Common.nList[zI].E[r, c]);
			}
		}

		public static void SubN(MultiM x, MultiM y, int zI) // z = x - y
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.SubN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
						XPCalLst.SetToNM(x.E[r, c] - y.E[r, c], Common.nList[zI].E[r, c]);
			}
		}

		public static void RAdd1N(int d1, int zI) // z = z + d1
		// allows literal d1, d1 is not changed, sets muErr true if overflow
		{
			MultiM t = new MultiM(2); // temp
			MultiFD.SetTo1F(d1, t);
			RAddN(t, zI);
			MultiFD.DoneF(t);
		}

		public static void Add1N(MultiM x, int d1, int zI) // z = x + d1
		// allows literal d1, d1 is not changed, sets muErr true if overflow
		{
			Calc.SetToN(x, zI);
			RAdd1N(d1, zI);
		}

		public static void FormatN(StreamWriter outF, string name, MultiM z)
		// output z as a formatted set of lines and pages
		// z is not modified
		{
			if (z.Rs == 0)
			{
				XPCalLst.FormatN(outF, name, z);
				return;
			}
			bool wB = MultiID.writeBool;
			MultiID.writeBool = false;
			int bO = MultiID.baseO;
			MultiID.baseO = 10;
			int fTn = MultiFD.fTn;
			MultiFD.fTn = 0;
			bool wE = MultiFD.writeExtra;
			MultiFD.writeExtra = false;
			string nLc = MultiID.nLc;
			MultiID.nLc = "-" + MultiID.nL;
			MultiID.WriteLn(null, outF, "@Echo Off");
			MultiID.WriteLn(null, outF, "Rem  Code file for matrix " + name);
			int rs = z.Rs;
			int cs = z.Cs;
			MultiID.WriteLn(null, outF);
			MultiID.WriteLn(null, outF, name + " = Mat(" + rs + ", " + cs + ");");
			for (int r = 0; r < rs; r++)
				for (int c = 0; c < cs; c++)
				{
					int tot = 0;
					MultiID.WriteLn(null, outF);
					MultiID.WriteT(null, outF, name + " = Set(" + name + ", " + (r+uFirst) + ", " + (c+uFirst) + ", ", ref tot);
					MultiFD.WriteF(null, outF, z.E[r, c], ref tot, 0);
					MultiID.WriteTLn(null, outF, ");", ref tot);
				}
			MultiID.writeBool = wB;
			MultiID.baseO = bO;
			MultiFD.fTn = fTn;
			MultiFD.writeExtra = wE;
			MultiID.nLc = nLc;
		} // FormatN

		public static void RepUnitTN(MultiM x, MultiM y, int zI) // z = x -- y
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.RepUnitTN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "RepUnitTN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.RepUnitTN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RepUnitTN

		public static void CompN(MultiM x, MultiM y, string op, string op2, int zI) // z = x op y
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.CompN(x, y, op, op2, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "CompN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.CompN(x.E[r, c], y.E[r, c], op, op2, tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // CompN

		public static void AndN(MultiM x, MultiM y, int zI) // z = x & y
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.AndN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "AndN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.AndN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // AndN

		public static void OrN(MultiM x, MultiM y, int zI) // z = x & y
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.OrN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "OrN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.OrN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // OrN

		public static void RMul1N(int d1, int zI) // z = z * d1
		// Multi-precision one or two super digit multiply d1 is not changed
		// muErr set true if z overflows
		{
			MultiM t = new MultiM(2);
			MultiMD.SetTo1M(d1, t);
			RMulN(t, zI);
			MultiFD.DoneF(t);
		}

		public static void Mul1N(MultiF x, int d1, int zI) // z = x * d1
		// Multi-precision one or two super digit multiply x and d1 are not changed
		// muErr set true if z overflows
		{
			Calc.SetToN(x, zI);
			RMul1N(d1, zI);
		}

		public static void MulN(MultiM x, MultiM y, int zI) // z = x * y
		// x or y or both may have the same address in memory as z
		// x, y are not changed unless they share memory with z
		// sets muErr true if overflow
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.MulN(x, y, zI);
			else if (x.Rs != 0 && y.Rs != 0 && x.Cs != y.Rs)
				MultiMD.ErrIncompatible();
			else if (x.Rs == 0)
			{
				MultiM z = new MultiM(Calc.fMC, y.Rs, y.Cs);
				MultiMD.MulM(x, y, z);
				Calc.SetToN(z, zI);
			}
			else if (y.Rs == 0)
			{
				MultiM z = new MultiM(Calc.fMC, x.Rs, x.Cs);
				MultiMD.MulM(x, y, z);
				Calc.SetToN(z, zI);
			}
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiM z = new MultiM(Calc.fMC, x.Rs, y.Cs);
				MultiMD.MulM(x, y, z);
				Calc.SetToN(z, zI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
			}
		} // MulN

		public static void MulQN(MultiM x, MultiM y, int zI) // z = x * y, all quaternions
		// x or y or both may have the same address in memory as z
		// x, y are not changed unless they share memory with z
		// sets muErr true if overflow
		{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiM z = new MultiM(Calc.fMC, 4, 1);
				MultiMD.MulQ(x, y, z);
				Calc.SetToN(z, zI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
		} // MulN

		public static void RMulN(MultiM x, int zI) // z = z * x
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow
		{
			MulN(Common.nList[zI], x, zI);
		}

		public static void SqN(MultiM x, int zI) // z = x * x
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow
		{
			MulN(x, x, zI);
		}

		public static void RSqN(int zI) // z = z * z
		// sets muErr true if overflow
		{
			int rs = Common.nList[zI].Rs;
			int cs = Common.nList[zI].Cs;
			if (rs == 0)
				XPCalLst.RSqN(zI);
			else if (rs != cs)
				MultiMD.ErrNotSquare();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiM z = new MultiM(Calc.fMC, rs, cs);
				z &= Common.nList[zI];
				MultiMD.RSqM(z);
				Calc.SetToN(z, zI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
			}		
		} // RSqN

		public static void PowMF(int yI, MultiM b, MultiM p,
			bool MFlag, int fMBI)
		// y = b^p % FMB (% only if MFlag is true)
		{
			if (b.Rs == 0 && p.Rs == 0)
				XPCalLst.PowMF(yI, b, p, MFlag, fMBI);
			else if (p.Rs != 0)
				MultiMD.ErrExponent();
			else
			{
				int tI;
				Common.InitNAdd(Common.nList[yI].N, "PowMF-tI", out tI);
				MultiM z = new MultiM(Calc.fMC, b.Rs, b.Cs);
				for (int r = 0; r < b.Rs; r++)
					for (int c = 0; c < b.Cs; c++)
					{
						XPCalLst.PowMF(tI, b.E[r, c], p, MFlag, fMBI);
						z.E[r, c] &= Common.nList[tI];
					}
				Calc.SetToN(z, yI);
				Common.LDelete(tI);
			}
		} // PowMF

		public static void PowN(MultiM x, MultiM y, int zI) // z = x^y
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.PowN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "PowN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.PowN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // PowN

		public static void RPowN(MultiM p, int zI) // z = z^p
		{
			PowN(Common.nList[zI], p, zI);
		}

		public static void PowMN(MultiM x, MultiM y, int zI) // z = x^y % FMB
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.PowMN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "PowMN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.PowMN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // PowMN

		public static void RPowMN(MultiM p, int zI) // z = z^p % FMB
		{
			PowMN(Common.nList[zI], p, zI);
		}

		public static void DivN(MultiM u, MultiM d, int zI) // z = u / d
		// u or d or both may have the same address in memory as z
		// u, d are not changed unless they share memory with z
		// sets muErr true if d = 0
		{
			MultiM z;
			int n = 1;
			if(d.Rs == 0)
			{
				z = new MultiM(Calc.fMC, u.Rs, u.Cs);
			}
			else
			{
				n = Math.Max(u.Rs, d.Rs);
				z = new MultiM(Calc.fMC, n, n);
			}
			MultiMD.DivM(u, d, z);
			Calc.SetToN(z, zI);
		} // DivN

		public static void DivRemN(MultiM x, MultiM y, int zI) // zij = Floor(xij / yij), Reij = remainder
		// u or d or both may have the same address in memory as z
		// u, d are not changed unless they share memory with z
		// sets muErr true if d = 0
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.DivRemN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int reI;
				Common.LFindAdd("Re", out reI);
				if (reI == zI)
				{
					MultiID.MuWriteErr(
					"Cannot set same location to both quotient and remainder, continuing...");
					return;
				}
				int tI;
				int reMI;
				Common.InitNAdd(1, "DivRemN-tI", out tI);
				Common.InitNAdd(1, "DivRemN-x2MI", out reMI);
				MatrixN(x.Rs, x.Cs, reMI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.DivRemN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(reMI);
							Common.LDelete(tI);
							Calc.SetToN(MultiMD.mZero, reI);
							Calc.SetToN(MultiMD.mZero, zI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
						XPCalLst.SetToNM((MultiF)Common.nList[reI], Common.nList[reMI].E[r, c]);
					}
				Calc.SetToII(reMI, reI);
				Common.LDelete(reMI);
				Common.LDelete(tI);
			}
		} // DivRemN

		public static void DivIntN(MultiM x, MultiM y, int zI) // zij = Floor(xij / yij)
		// u or d or both may have the same address in memory as z
		// u, d are not changed unless they share memory with z
		// sets muErr true if d = 0
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.DivIntN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "DivIntN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.DivIntN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // DivIntN

		public static void RDivN(MultiM d, int zI) // z = z / d
		// d may have the same address in memory as z
		// d is not changed unless it shares memory with z
		// sets muErr true if d = 0
		{
			DivN(Common.nList[zI], d, zI);
		}

		public static void FacSN(MultiM x, int zI, bool MFlag, int fMBI)
		// z = x! = 1*2*3*...*x % FMB
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XPCalLst.FacSN(x, zI, MFlag, fMBI);
			else
			{
				int tI;
				Common.InitNAdd(1, "FacN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.FacSN(x.E[r, c], tI, MFlag, fMBI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // FacSN

		public static void Fac2N(MultiM x, MultiM y, int zI, bool MFlag)
		// z = Fac(x) / Fac(y) % fMB by binary splitting
		// x or y or both may have the same address in memory as z
		// x, y are not changed unless they share memory with z
		// sets muErr true if y = 0
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.Fac2N(x, y, zI, MFlag);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "Fac2N-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.Fac2N(x.E[r, c], y.E[r, c], tI, MFlag);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // Fac2N

		public static void FacN(MultiM x, int zI, bool MFlag, int fMBI)
		// z = x! = 1*2*3*...*x % FMB
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XPCalLst.FacN(x, zI, MFlag, fMBI);
			else
			{
				int tI;
				Common.InitNAdd(1, "FacN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.FacN(x.E[r, c], tI, MFlag, fMBI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // FacN

		public static void AFibN(MultiM x, int zI)
		// z = Arc Fibonacci number Abs(x), always an integer, accurate if x is a Fibonacci number
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XPCalLst.AFibN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "AFibN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.AFibN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // AFibN

		public static void FibN(MultiM x, int zI)
		// z = Fibonacci number of Int(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XPCalLst.FibN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "FibN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.FibN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // FibN

		public static void RIsFibN(int zI)
		// z = 1 (True) if |z| is a Fibonacci number, else 0 (False)
		// True iff z is an integer and the closed interval
		// [Phi*z - 1/z, Phi*z + 1/z] contains a positive integer
		{
			if (Common.nList[zI].Rs != 0)
				MultiMD.ErrNotScalar();
			else
				XPCalLst.RIsFibN(zI);
		} // RIsFibN

		public static void RIsFib2N(int zI)
		// z = 1 (True) if |z| is a Fibonacci number, else 0 (False)
		// True iff z is an integer and 5z^2+4 or 5z^2-4 is a perfect square
		{
			if (Common.nList[zI].Rs != 0)
				MultiMD.ErrNotScalar();
			else
				XPCalLst.RIsFib2N(zI);
		} // RIsFib2N

		public static void Lerch1(MultiM x, MultiM s, MultiM a, bool T, int zI) // z = Lerch transcendent capitol Phi(x, s, a)
		// x, s, a may have the same address in memory as z				// to 16 decimal digits and traditional
		// x, s, a are not changed unless it shares memory with x
		// sets muErr true if overflow or error
		//
		// The LerchT1(x, s, a) function is Psi(x, s, a) = Sum {k = 0, 1, ...} of x^k / (a + k)^s,
		// where |x| <= 1 and a != 0, -1, -2, ... .
		// Lerch1(x, s, a) = sum[k=0, infinity]{x^k / |a + k|^s}, terms with a + k == 0 are ignored, |x| <= 1
		{
			if (x.Rs == 0 && s.Rs == 0 && a.Rs == 0)
				XPCalLst.Lerch1(x, s, a, T, zI);
			else if (x.Rs != s.Rs || x.Rs != a.Rs || x.Cs != s.Cs || x.Cs != a.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "Lerch1-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.Lerch1(x.E[r, c], s.E[r, c], a.E[r, c], T, tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // Lerch1

		public static void Lerch(MultiM x, MultiM s, MultiM a, bool T, int zI) // z = Lerch transcendent capitol Phi(x, s, a)
		// Function that computes approximation to Lerch Phi as a
		// converging sequence of CNC transforms S^n_k.
		// x, s, a may have the same address in memory as z				// traditional
		// x, s, a are not changed unless it shares memory with z
		// sets muErr true if overflow or error
		//
		// The LerchT(x, s, a) function is Phi(x, s, a) = Sum {k = 0, 1, ...} of x^k / (a + k)^s,
		// where |x| <= 1 and a != 0, -1, -2, ... .
		// Lerch(x, s, a) = sum[k=0, infinity]{x^k / |a + k|^s}, terms with a + k == 0 are ignored, |x| <= 1
		{
			if (x.Rs == 0 && s.Rs == 0 && a.Rs == 0)
				XPCalLst.Lerch(x, s, a, T, zI);
			else if (x.Rs != s.Rs || x.Rs != a.Rs || x.Cs != s.Cs || x.Cs != a.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "Lerch-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.Lerch(x.E[r, c], s.E[r, c], a.E[r, c], T, tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // Lerch

		public static void Lerch2(MultiM x, MultiM s, MultiM a, bool T, int zI) // z = Lerch transcendent capitol Phi(x, s, a)
		// x, s, a may have the same address in memory as z			// by a simple sum and traditional
		// x, s, a are not changed unless it shares memory with z
		// sets muErr true if overflow or error
		//
		// The LerchT2(x, s, a) function is Phi(x, s, a) = Sum {k = 0, 1, ...} of x^k / (a + k)^s,
		// where |x| <= 1 and a != 0, -1, -2, ... .
		// Lerch2(x, s, a) = sum[k=0, infinity]{x^k / |a + k|^s}, terms with a + k == 0 are ignored, |x| <= 1
		{
			if (x.Rs == 0 && s.Rs == 0 && a.Rs == 0)
				XPCalLst.Lerch2(x, s, a, T, zI);
			else if (x.Rs != s.Rs || x.Rs != a.Rs || x.Cs != s.Cs || x.Cs != a.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "Lerch2-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.Lerch2(x.E[r, c], s.E[r, c], a.E[r, c], T, tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // Lerch2

		public static void RPYMN(MultiM r, MultiM p, MultiM y, int zI) // z = Rotation matrix
		// r = roll, p = pitch, y = yaw
		// roll matrix = {1, 0, 0; 0, cos(r), sin(r); 0, -sin(r), cos(r)}
		// pitch matrix = {cos(p), 0, -sin(p); 0, 1, 0; sin(p), 0, cos(p)}
		// yaw matrix = {cos(y), sin(y), 0; 0, -sin(y), cos(y), 0; 0, 0, 1}
		// Rotation matrix = yaw * pitch * roll
		{
			int tI;
			Common.InitNAdd(1, "RPYMN-tI", out tI);

			int rI;
			Common.InitNAdd(1, "RPYMN-rI", out rI);
			MatrixN(3, 3, rI);
			CosN(r, tI);
			XPCalLst.SetToNM(Common.nList[tI], Common.nList[rI].E[1, 1]);
			SinN(r, tI);
			XPCalLst.SetToNM(Common.nList[tI], Common.nList[rI].E[1, 2]);
			XPCalLst.SetToNM(-Common.nList[rI].E[1, 2], Common.nList[rI].E[2, 1]);
			XPCalLst.SetToNM(Common.nList[rI].E[1, 1], Common.nList[rI].E[2, 2]);
			XPCalLst.SetToNM(MultiFD.fOne, Common.nList[rI].E[0, 0]);

			int pI;
			Common.InitNAdd(1, "RPYMN-pI", out pI);
			MatrixN(3, 3, pI);
			CosN(p, tI);
			XPCalLst.SetToNM(Common.nList[tI], Common.nList[pI].E[0, 0]);
			SinN(p, tI);
			XPCalLst.SetToNM(Common.nList[tI], Common.nList[pI].E[2, 0]);
			XPCalLst.SetToNM(-Common.nList[pI].E[2, 0], Common.nList[pI].E[0, 2]);
			XPCalLst.SetToNM(Common.nList[pI].E[0, 0], Common.nList[pI].E[2, 2]);
			XPCalLst.SetToNM(MultiFD.fOne, Common.nList[pI].E[1, 1]);

			int yI;
			Common.InitNAdd(1, "RPYMN-yI", out yI);
			MatrixN(3, 3, yI);
			CosN(y, tI);
			XPCalLst.SetToNM(Common.nList[tI], Common.nList[yI].E[0, 0]);
			SinN(y, tI);
			XPCalLst.SetToNM(Common.nList[tI], Common.nList[yI].E[0, 1]);
			XPCalLst.SetToNM(-Common.nList[yI].E[0, 1], Common.nList[yI].E[1, 0]);
			XPCalLst.SetToNM(Common.nList[yI].E[0, 0], Common.nList[yI].E[1, 1]);
			XPCalLst.SetToNM(MultiFD.fOne, Common.nList[yI].E[2, 2]);

			MulN(Common.nList[yI], Common.nList[pI], yI);
			MulN(Common.nList[yI], Common.nList[rI], yI);

			Calc.SetToII(yI, zI);
			Common.LDelete(yI);
			Common.LDelete(pI);
			Common.LDelete(rI);
			Common.LDelete(tI);
		} // RPYMN

		public static void GamN(MultiM x, int zI)
		// z = gamma(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XPCalLst.GamN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "GamN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
				MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
				for (int c = 0; c < x.Cs; c++)
				{
					XPCalLst.GamN(x.E[r, c], tI);
					if (MultiID.muErr)
					{
						Common.LDelete(tI);
						return;
					}
					XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
				}
				Common.LDelete(tI);
			}		
		} // GamN

		public static void BetaN(MultiM x, int zI)
		// z = beta(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XPCalLst.BetaN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "BetaN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.BetaN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // BetaN

		public static void BetaL(MultiM x, int zI)
		// z = beta(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XPCalLst.BetaL(x, Common.nList[zI]);
			else
			{
				int tI;
				Common.InitNAdd(1, "BetaL-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.BetaL(x.E[r, c], Common.nList[tI]);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // BetaL

		public static void ErfN(MultiM x, int zI)
		// z = error function erf(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XPCalLst.ErfN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ErfN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.ErfN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ErfN

		public static void ErfCN(MultiM x, int zI)
		// z = complementary error function erfC(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XPCalLst.ErfCN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ErfCN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.ErfCN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ErfCN

		public static void EtaN(MultiM x, int zI)
		// z = eta(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XPCalLst.EtaN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "EtaN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.EtaN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // EtaN

		public static void LambdaN(MultiM x, int zI)
		// z = lambda(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XPCalLst.LambdaN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "LambdaN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.LambdaN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // LambdaN

		public static void ZetaN(MultiM x, int zI)
		// z = Zeta(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XPCalLst.ZetaN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ZetaN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.ZetaN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // ZetaN

		public static void GamN1(MultiM x, int zI) // 16-digit gamma function
		// z = gamma(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XPCalLst.GamN1(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "GamN1-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.GamN1(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // GamN1

		public static void PsiN(MultiM x, int zI)
		// z = psi(x) = gamma'(x) / gamma(x) = digamma function
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XPCalLst.PsiN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "PsiN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.PsiN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // PsiN

		public static void RFacN(int zI) // z = z! = 1*2*3*...*x
		// sets muErr true if overflow or error
		{
			int fMBI = 1;
			FacN(Common.nList[zI], zI, false, fMBI);
		}

		public static void RFacSN(int zI) // z = z! = 1*2*3*...*x
		// sets muErr true if overflow or error
		{
			int fMBI = 1;
			FacSN(Common.nList[zI], zI, false, fMBI);
		}
		public static void RAFibN(int zI) // z = Arc Fibonacci number Abs(z)
		// sets muErr true if overflow or error
		{
			AFibN(Common.nList[zI], zI);
		}

		public static void RFibN(int zI) // z = Fibonacci number Int(z)
		// sets muErr true if overflow or error
		{
			FibN(Common.nList[zI], zI);
		}

		public static void RFacMBN(int zI) // z = z! = 1*2*3*...*x % FMB
		// sets muErr true if overflow or error
		{
			int fMBI;
			if (Common.FMBIok(zI, out fMBI))
				FacN(Common.nList[zI], zI, true, fMBI);
		}

		public static void RFacMBSN(int zI) // z = z! = 1*2*3*...*x % FMB
		// sets muErr true if overflow or error
		{
			int fMBI;
			if (Common.FMBIok(zI, out fMBI))
				FacSN(Common.nList[zI], zI, true, fMBI);
		}

		public static void RBetaN(int zI) // z = beta(z)
		// sets muErr true if overflow or error
		{
			BetaN(Common.nList[zI], zI);
		}

		public static void RErfN(int zI) // z = erf(z)
		// sets muErr true if overflow or error
		{
			ErfN(Common.nList[zI], zI);
		}

		public static void RErfCN(int zI) // z = erfc(z)
		// sets muErr true if overflow or error
		{
			ErfCN(Common.nList[zI], zI);
		}

		public static void REtaN(int zI) // z = eta(z)
		// sets muErr true if overflow or error
		{
			EtaN(Common.nList[zI], zI);
		}

		public static void RLambdaN(int zI) // z = lambda(z)
		// sets muErr true if overflow or error
		{
			LambdaN(Common.nList[zI], zI);
		}

		public static void RZetaN(int zI) // z = Zeta(z)
		// sets muErr true if overflow or error
		{
			ZetaN(Common.nList[zI], zI);
		}

		public static void RGamN(int zI) // z = gamma(z)
		// sets muErr true if overflow or error
		{
			GamN(Common.nList[zI], zI);
		}

		public static void RGamN1(int zI) // z = gamma(z), 16-digit gamma function
		// sets muErr true if overflow or error
		{
			GamN1(Common.nList[zI], zI);
		}

		public static void RPsiN(int zI) // z = psi(z)
		// sets muErr true if overflow or error
		{
			PsiN(Common.nList[zI], zI);
		}

		public static void SqRootN(MultiM x, int zI) // z = SqRt(x)
		// x is not changed, it is ok for x and z to have the same location in memory
		// but then x will be changed. Sets muErr true if error
		{
			if (x.Rs == 0)
				XPCalLst.SqRootN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "SqRootN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.SqRootN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // SqRootN

		public static void RCuRtN(int zI) // z = SqRt(z)
		// sets muErr true if error
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XPCalLst.RCuRtN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RCuRtN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						XPCalLst.SetToNM(Common.nList[zI].E[r, c], (MultiF)Common.nList[tI]);
						XPCalLst.RCuRtN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RCuRtN

		public static void RSqRtN(int zI) // z = SqRt(z)
		// sets muErr true if error
		{
			SqRootN(Common.nList[zI], zI);
		}

		public static void RSqRtRemN(int zI) // z = Floor(SquareRoot(z)) and set Re to remainder
		// sets muErr true if error
		{
			if (Common.nList[zI].Rs == 0)
				XPCalLst.RSqRtRemN(zI);
			else
			{
				int reI;
				Common.LFindAdd("Re", out reI);
				if (reI == zI)
				{
					MultiID.MuWriteErr(
					"Cannot set same location to both SqRt and remainder, continuing...");
					return;
				}
				int tI;
				int reMI;
				Common.InitNAdd(1, "RSqRtRemN-tI", out tI);
				Common.InitNAdd(1, "RSqRtRemN-reMI", out reMI);
				MatrixN(Common.nList[zI].Rs, Common.nList[zI].Cs, reMI);
				for (int r = 0; r < Common.nList[zI].Rs; r++)
					for (int c = 0; c < Common.nList[zI].Cs; c++)
					{
						Calc.SetToN(Common.nList[zI].E[r, c], tI);
						XPCalLst.RSqRtRemN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(reMI);
							Common.LDelete(tI);
							Calc.SetToN(MultiMD.mZero, reI);
							Calc.SetToN(MultiMD.mZero, zI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
						XPCalLst.SetToNM((MultiF)Common.nList[reI], Common.nList[reMI].E[r, c]);
					}
				Calc.SetToII(reMI, reI);
				Common.LDelete(reMI);
				Common.LDelete(tI);
			}
		} // RSqRtRemN

		public static void SolveN(MultiM x, MultiM y, int zI) // solve for z, x * z = y
		// this code is adapted from Numerical Recipes in C
		{
			if (x.Rs != x.Cs || y.Rs != x.Rs || y.Cs != 1)
				MultiMD.ErrIncompatible();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int n = x.Rs;
				MultiM a = new MultiM(Calc.fMC, n, n);
				a &= x;
				int[] index;
				index = new int[n];
				MultiF d = new MultiF();
				bool sing;
				MultiMD.LUdecomp(a, index, d, out sing); // decompose the matrix
				if (sing)
					MultiMD.ErrSingular();
				else
				{
					MultiM b = new MultiM(Calc.fMC, n, 1);
					b &= y;
					MultiMD.LUbackSub(a, index, b);
					Calc.SetToN(b, zI);
				}
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
			}
		} // SolveN

		public static void SolveSN(MultiM x, MultiM y, int zI) // solve for z, x * z = y
		// this code is adapted from Numerical Recipes in C
		{
			if (y.Rs != x.Rs || y.Cs != 1 || x.Rs == 0)
				MultiMD.ErrIncompatible();
			else if (x.Rs < x.Cs)
			{
				MultiM xx = new MultiM(1, x.Cs, x.Cs);
				MultiM yy = new MultiM(1, x.Cs, 1);
				for (int r = 0; r < x.Rs; r++)
				{
					XPCalLst.SetToNM(y.E[r, 0], yy.E[r, 0]);
					for (int c = 0; c < x.Cs; c++)
						XPCalLst.SetToNM(x.E[r, c], xx.E[r, c]);
				}
				for (int r = x.Rs; r < x.Cs; r++)
				{
					yy.E[r, 0] &= MultiFD.fZero;
					for (int c = 0; c < x.Cs; c++)
						xx.E[r, c] &= MultiFD.fZero;
				}
				SolveSSN(xx, yy, zI);
			}
			else if (x.Rs > x.Cs)
			{
				int rs = x.Rs;
				int cs = x.Cs;
				MultiM xx = new MultiM(Calc.fMC, x.Rs, x.Rs);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
						XPCalLst.SetToNM(x.E[r, c], xx.E[r, c]);
				for (int r = 0; r < x.Rs; r++)
					for (int c = x.Cs; c < x.Rs; c++)
						xx.E[r, c] &= MultiFD.fZero;
				SolveSSN(xx, y, zI);
				MultiM yy = new MultiM(Calc.fMC, cs, 1);
				for (int r = 0; r < cs; r++)
					yy.E[r, 0] &= Common.nList[zI].E[r, 0];
				Calc.SetToN(yy, zI);
			}
			else
				SolveSSN(x, y, zI);
		} // SolveN

		public static void SolveSSN(MultiM x, MultiM y, int zI) // solve for z, x * z = y, x square
		// this code is adapted from Numerical Recipes in C
		{
			if (x.Rs != x.Cs || y.Rs != x.Rs || y.Cs != 1 || x.Rs == 0)
				MultiMD.ErrIncompatible();
			else
			{
				int i;
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int n = x.Rs;
				MultiM u = new MultiM(Calc.fMC, n, n);
				MultiM w = new MultiM(Calc.fMC, 1, n);
				MultiM v = new MultiM(Calc.fMC, n, n);
				u &= x;
				SVD(u, w, v);
				string sing = "";
				int rank = 0;
				for (i = 0; i < n; i++)
					if (w.E[0, i] != MultiFD.fZero)
					{
						w.E[0, i] &= MultiFD.fOne / w.E[0, i];
						rank++;
					}
					else
						sing = "Singular, ";
				MultiM z = new MultiM(Calc.fMC, n, 1);
				MultiM r = new MultiM(Calc.fMC, n, 1);
				MultiMD.SetToM(v * (MultiMD.Diagonal(w) * (Transpose(u) * y)), z);
				r &= x * z - y;
				long next;
				long max = Int64.MinValue;
				for (i = 0; i < n; i++)
				{
					if (y.E[i, 0] != MultiFD.fZero)
					{
						next = y.E[i, 0].N + y.E[i, 0].C;
						if (next > max)
							max = next;
					}
				}
				for (i = 0; i < n; i++)
				{
					next = r.E[i, 0].N + r.E[i, 0].C;
					if (max - next > Calc.fMC - MultiFD.fGuard)
						MultiFD.ClearF(r.E[i, 0]);
				}
				for (i = 0; i < n; i++)
					r.E[0, 0] += r.E[i, 0] * r.E[i, 0];
				MultiFD.SqRootF(r.E[0, 0], r.E[0, 0]); 
				max = Int64.MinValue;
				for (i = 0; i < n; i++)
				{
					if (z.E[i, 0] != MultiFD.fZero)
					{
						next = z.E[i, 0].N + z.E[i, 0].C;
						if (next > max)
							max = next;
					}
				}
				for (i = 0; i < n; i++)
				{
					next = z.E[i, 0].N + z.E[i, 0].C;
					if (max - next > Calc.fMC - MultiFD.fGuard)
						MultiFD.ClearF(z.E[i, 0]);
				}
				next = r.E[0, 0].N + r.E[0, 0].C;
				if (max - next > Calc.fMC - MultiFD.fGuard)
					MultiFD.ClearF(r.E[0, 0]);
				Calc.SetToN(z, zI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				MultiFD.NormF(r.E[0, 0]);
				int tot = 0;
				MultiID.WriteT(sing + "Rank = " + rank + ", Residual = ", ref tot);
				MultiFD.WriteF(MultiID.outS, null, r.E[0, 0], ref tot, 0);
				Calc.NormN(zI);
			}
		} // SolveSSN

		public static void SolveH(MultiM x, MultiM z) // solves homogeneous equation x * z = 0
		// z = all non trivial solutions of matrix x * z = 0
		{
			int m = x.Rs;
			int n = x.Cs;
			if (m == 0)
				MultiMD.ErrNotMatrix();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int j, ri, ci;
				MultiF di = new MultiF();
				if (m < n)
					m = n;
				MultiM u = new MultiM(Calc.fMC, m, n);
				MultiM w = new MultiM(Calc.fMC, 1, n);
				MultiM v = new MultiM(Calc.fMC, n, n);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < n; c++)
						MultiFD.SetToF(x.E[r, c], u.E[r, c]);
				SVD(u, w, v);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormNM(w);
				XPCalLst.Expand();
				for (j = n - 1; j >= 0; j--)			// delete columns of v where w != 0
				{
					if (w.E[0, j] != MultiFD.fZero)
						Del(v, j, v);
				}
				MultiM zz = new MultiM(1, n, v.Cs);
				for (ci = 0; ci < v.Cs; ci++)			// copy v to zz
					for (ri = 0; ri < n; ri++)
						XPCalLst.SetToNM(v.E[ri, ci], zz.E[ri, ci]);
				Calc.fMC -= MultiFD.fGuard;
				Calc.NormNMCol(zz);
				Calc.fMC += MultiFD.fGuard;
				bool err = false;
				for (ci = zz.Cs -1; ci >= 0; ci--)		// for all columns of zz
				{
					XPCalLst.SetToNM(MultiFD.fOne, di);
					for (ri = n - 1; ri >= 0; ri--)		// find last non-zero element in this column
						if (zz.E[ri, ci] != MultiFD.fZero)
						{
							XPCalLst.SetToNM(zz.E[ri, ci], di);
							break;
						}
					if (di == MultiFD.fZero)			// if column is all zeros
					{
						Del(zz, ci, zz);				// Delete columns of zz with all zeros
						err = true;
					}
					else
					{
						for (ri = 0; ri < n; ri++)		// Make last non-zero element equal to one
							XPCalLst.SetToNM(zz.E[ri, ci] / di, zz.E[ri, ci]);
					}
				}
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormNM(zz);
				for (ci = zz.Cs - 1; ci > 0; ci--)
				{
					m = 0;
					for (ri = 0; ri < n; ri++)
					{
						if (zz.E[ri, ci] == zz.E[ri, ci - 1])
							m++;
						else
							break;
					}
					if (m == n)
					{
						Del(zz, ci, zz);
						err = true;
					}
				}
				Calc.SetToNM(zz, z);
				if (err)
					MultiID.WriteLn("Not all vectors found");
			}
		} // SolveH

		public static void SolveHN(MultiM x, int zI) // solves homogeneous equation x * z = 0
		// z = all non trivial solutions of matrix x * z = 0
		{
			if (x.Rs == 0)
				MultiMD.ErrNotMatrix();
			else
			{
				MultiM z = new MultiM();
				SolveH(x, z);
				Calc.SetToN(z, zI);
			}
		} // SolveHN

		public static int Get1InM(int low, int high, MultiM z)
		{
			if (z.Rs != 0) return low;
			int d1;
			if (z.N + z.C > 1)
			{
				d1 = 100000000;
				if (z.S) d1 = -d1;
			}
			else
				MultiFD.Get1F(out d1, z);
			if (d1 < low) return low;
			if (d1 > high) return high;
			return d1;
		} // Get1InM

		public static void SortC(MultiM x) // x = sorted columns of x
		{
			MultiM rightM = new MultiM(2);
			MultiFD.SetTo1F(x.Rs, rightM);
			SortCS(x, MultiMD.mOne, rightM);
		} // SortC

		public static void SortCS(MultiM x, MultiM leftM, MultiM rightM) // x = sorted columns of x
		{
			swapped = false;
			if (x.Rs == 0)
				return;
			int left = Get1InM(1, x.Rs, leftM);
			left -= uFirst;
			int right = Get1InM(1, x.Rs, rightM);
			if (rightM.Rs != 0)
				right = x.Rs;
			right -= uFirst;
			if (left > right)
			{
				int swap = left;
				left = right;
				right = swap;
			}
			for (int c = 0; c < x.Cs; c++)
				QuickSortC(x, left, right, c);
			if (swapped && !Common.quiet) MultiID.WriteLn("SortC: Some elements were swapped");
			if (!swapped && !Common.quiet) MultiID.WriteLn("SortC: No elements were swapped");
		} // SortCS

		public static void QuickSortC(MultiM x, int left, int right, int i) // sort column i of x
		{
			if (right <= left)
				return;
			int leftIndex = left;
			int rightIndex = right;
			// Arbitrarily establishing partition element as the midpoint of the array.
			int j = (left + right) / 2;
			MultiF partionElement = new MultiF(x.E[j, i].M);
			partionElement &= x.E[j, i];
			// loop through the array until indices cross
			while (leftIndex <= rightIndex)
			{
				// find the first element that is greater than or equal to
				// the partionElement starting from the leftIndex.
				while ((leftIndex < right) && (x.E[(leftIndex), i] < partionElement))
					leftIndex = leftIndex + 1;
				// find an element that is smaller than or equal to
				// the partionElement starting from the rightIndex.
				while ((rightIndex > left) && (x.E[(rightIndex), i] > partionElement))
					rightIndex = rightIndex - 1;
				// if the indexes have not crossed, swap
				if (leftIndex <= rightIndex)
				{
					if (leftIndex != rightIndex)
					{
						SwapC(ref x.E[(leftIndex), i], ref x.E[(rightIndex), i]);
						swapped = true;
					}
					leftIndex = leftIndex + 1;
					rightIndex = rightIndex - 1;
				}
			}
			// If the right index has not reached the left side of array
			// must now sort the left partition.
			if (left < rightIndex)
				QuickSortC(x, left, rightIndex, i);
			// If the left index has not reached the right side of array
			// must now sort the right partition.
			if (leftIndex < right)
				QuickSortC(x, leftIndex, right, i);
		} // QuickSortC

		public static void SortR(MultiM x) // x = sorted rows of x
		{
			MultiM rightM = new MultiM(2);
			MultiFD.SetTo1F(x.Cs, rightM);
			SortRS(x, MultiMD.mOne, rightM);
		} // SortR

		public static void SortRS(MultiM x, MultiM leftM, MultiM rightM) // x = sorted rows of x
		{
			swapped = false;
			if (x.Rs == 0)
				return;
			int left = Get1InM(1, x.Cs, leftM);
			left -= uFirst;
			int right = Get1InM(1, x.Cs, rightM);
			if (rightM.Rs != 0)
				right = x.Cs;
			right -= uFirst;
			if (left > right)
			{
				int swap = left;
				left = right;
				right = swap;
			}
			for (int r = 0; r < x.Rs; r++)
				QuickSortR(x, left, right, r);
			if (swapped && !Common.quiet) MultiID.WriteLn("SortR: Some elements were swapped");
			if (!swapped && !Common.quiet) MultiID.WriteLn("SortR: No elements were swapped");
		} // SortRS

		public static void QuickSortR(MultiM x, int left, int right, int i) // sort row i of x
		{
			if (right <= left)
				return;
			int leftIndex = left;
			int rightIndex = right;
			// Arbitrarily establishing partition element as the midpoint of the array.
			int j = (left + right) / 2;
			MultiF partionElement = new MultiF(x.E[i, j].M);
			partionElement &= x.E[i, j];
			// loop through the array until indices cross
			while (leftIndex <= rightIndex)
			{
				// find the first element that is greater than or equal to
				// the partionElement starting from the leftIndex.
				while ((leftIndex < right) && (x.E[i, (leftIndex)] < partionElement))
					leftIndex = leftIndex + 1;
				// find an element that is smaller than or equal to
				// the partionElement starting from the rightIndex.
				while ((rightIndex > left) && (x.E[i, (rightIndex)] > partionElement))
					rightIndex = rightIndex - 1;
				// if the indexes have not crossed, swap
				if (leftIndex <= rightIndex)
				{
					if (leftIndex != rightIndex)
					{
						SwapC(ref x.E[i, (leftIndex)], ref x.E[i, (rightIndex)]);
						swapped = true;
					}
					leftIndex = leftIndex + 1;
					rightIndex = rightIndex - 1;
				}
			}
			// If the right index has not reached the left side of array
			// must now sort the left partition.
			if (left < rightIndex)
				QuickSortR(x, left, rightIndex, i);
			// If the left index has not reached the right side of array
			// must now sort the right partition.
			if (leftIndex < right)
				QuickSortR(x, leftIndex, right, i);
		} // QuickSortR

		public static void Image(MultiM x, MultiM z) // z = a basis for the image/range of matrix x
		{
			int m = x.Rs;
			int n = x.Cs;
			if (m == 0)
				MultiMD.ErrNotMatrix();
			else
			{
				MultiM xx = new MultiM();
				MultiM zz = new MultiM(1, m, n);
				RREF(x, xx);
				int c = 0;
				int ci = 0;
				for (int r = 0; r < m; r++)
				{
					for ( ; c < n; c++)
					{
						if (xx.E[r, c] == MultiFD.fOne)
						{
							// insert column c of x into column ci of zz
							for (int ri = 0; ri < m; ri++)
								XPCalLst.SetToNM(x.E[ri, c], zz.E[ri, ci]);
							ci++;
							c++;
							break;
						}
					}
				}
				zz.Cs = ci;
				Calc.SetToNM(zz, z);
			}
		} // Image

		public static void ImageN(MultiM x, int zI) // z = a basis for the image/range of matrix x
		{
			if (x.Rs == 0)
				MultiMD.ErrNotMatrix();
			else
			{
				MultiM z = new MultiM();
				Image(x, z);
				Calc.SetToN(z, zI);
			}
		} // ImageN

		public static void RankN(MultiM x, int zI) // z = rank(x)
		{
			int m = x.Rs;
			int n = x.Cs;
			if (m == 0)
				MultiMD.ErrNotMatrix();
			else
			{
				int rank = m;
				bool zero = true;
				MultiM z = new MultiM();
				RREF(x, z);
				for (int r = m - 1; r >= 0; r--)
				{
					for (int c = 0; c < n; c++)
					{
						if (z.E[r, c] != MultiFD.fZero)
						{
							zero = false;
							break;
						}
					}
					if (zero)
						rank--;
					else
						break;
				}
				Calc.ClearN(zI);
				Common.nList[zI].V[0] = rank;
			} 
		} // RankN

		public static MultiF Abs(MultiF x)	// return |x|
		{
			if (x < MultiFD.fZero)
				return -x;
			else
				return x;
		}

		public static MultiF Max(MultiF x, MultiF y)	// return larger of x and y
		{
			if (x >= y)
				return x;
			else
				return y;
		}

		public static MultiF Sign(MultiF x, MultiF y)	// return x with sign of y
		// actually, change sign of x if y < 0
		{
			if (x == MultiFD.fZero || y >= MultiFD.fZero)
				return x;
			else
				return -x;
		}
		public static MultiF Pythag(MultiF x, MultiF y)	// return SqRt(x^2 + y^2)
		{
			if (x == MultiFD.fZero)
				return Abs(y);
			if (y == MultiFD.fZero)
				return Abs(x);
			MultiF t = new MultiF(Calc.fMC);
			t = x * x + y * y;
			MultiFD.SqRootF(t, t);
			return t;
		}

		public static void SortUWV(MultiM u, MultiM w, MultiM v)
		{
			int m = u.Rs;
			int n = u.Cs;
			int i, j, k;
			MultiF x = new MultiF(1);
			MultiM uc = new MultiM(1, m, 1);
			MultiM vc = new MultiM(1, n, 1);
			for (j = 1; j < n; j++) // Sort w by straight insertion
			{
				XPCalLst.SetToNM(w.E[0, j], x);
				for (k = 0; k < m; k++)
					XPCalLst.SetToNM(u.E[k, j], uc.E[k, 0]);
				for (k = 0; k < n; k++)
					XPCalLst.SetToNM(v.E[k, j], vc.E[k, 0]);
				for (i = j - 1; i >= 0; i--)
				{
					if (w.E[0, i] <= x) // least first
						break;
					XPCalLst.SetToNM(w.E[0, i], w.E[0, i + 1]);
					for (k = 0; k < m; k++)
						XPCalLst.SetToNM(u.E[k, i], u.E[k, i + 1]);
					for (k = 0; k < n; k++)
						XPCalLst.SetToNM(v.E[k, i], v.E[k, i + 1]);
				}
				XPCalLst.SetToNM(x, w.E[0, i + 1]);
				for (k = 0; k < m; k++)
					XPCalLst.SetToNM(uc.E[k, 0], u.E[k, i + 1]);
				for (k = 0; k < n; k++)
					XPCalLst.SetToNM(vc.E[k, 0], v.E[k, i + 1]);
			}
		} // SortUWV

		public static void SVD(MultiM a, MultiM w, MultiM v)
		// Singular Value Decomposition, a = u * w * Tran(v), a replaced with u
		// the diagonal matrix is output as a row vector, a.Rs must be >= a.Cs
		// if smaller, a should be filled with zero rows
		// this code is adapted from Numerical Recipes in C
		{
			int m = a.Rs;
			int n = a.Cs;
			if (m == 0)
				MultiMD.ErrNotMatrix();
			else
			{
				int flag, i, its, j, jj, k, L = 0, nm = 0;
				MultiF anorm = new MultiF(Calc.fMC);
				MultiF c = new MultiF(Calc.fMC);
				MultiF f = new MultiF(Calc.fMC);
				MultiF g = new MultiF(Calc.fMC);
				MultiF h = new MultiF(Calc.fMC);
				MultiF s = new MultiF(Calc.fMC);
				MultiF scale = new MultiF(Calc.fMC);
				MultiF x = new MultiF(Calc.fMC);
				MultiF y = new MultiF(Calc.fMC);
				MultiF z = new MultiF(Calc.fMC);
				MultiM rv1 = new MultiM(Calc.fMC, 1, n);
				// Householder reduction to bidiagonal form
				g &= scale &= anorm &= MultiFD.fZero;
				for (i = 0; i < n; i++) 
				{
					L = i + 1;
					rv1.E[0, i] &= scale * g;
					g &= s &= scale &= MultiFD.fZero;
					if (i < m)
					{
						for (k = i; k < m; k++)
							scale += Abs(a.E[k, i]);
						if (scale != MultiFD.fZero)
						{
							for (k = i; k < m; k++)
							{
								a.E[k, i] /= scale;
								s += a.E[k, i] * a.E[k, i];
							}
							f &= a.E[i, i];
							MultiFD.SqRootF(s, g);	// g = -SIGN(sqrt(s), f);
							g &= -Sign(g, f);
							h &= f * g - s;
							a.E[i, i] &= f - g;
							for (j = L; j < n; j++)
							{
								for (s &= MultiFD.fZero, k = i; k < m; k++)
									s += a.E[k, i]* a.E[k, j];
								f &= s / h;
								for (k = i; k < m; k++)
									a.E[k, j] += f * a.E[k, i];
							}
							for (k = i; k < m; k++)
								a.E[k, i] *= scale;
						}
					}
					w.E[0, i] &= scale * g;
					g &= s &= scale &= MultiFD.fZero;
					if (i < m && i != n - 1)
					{
						for (k = L; k < n; k++)
							scale += Abs(a.E[i, k]);
						if (scale != MultiFD.fZero)
						{
							for (k = L; k < n; k++)
							{
								a.E[i, k] /= scale;
								s += a.E[i, k] * a.E[i, k];
							}
							f &= a.E[i, L];
							MultiFD.SqRootF(s, g);
							g &= -Sign(g, f);
							h &= f * g - s;
							a.E[i, L] &= f - g;
							for (k = L; k < n; k++)
								rv1.E[0, k] &= a.E[i, k] / h;
							for (j = L; j < m; j++)
							{
								for (s &= MultiFD.fZero, k = L; k < n; k++)
									s += a.E[j, k] * a.E[i, k];
								for (k = L; k < n; k++)
									a.E[j, k] += s * rv1.E[0, k];
							}
							for (k = L; k < n; k++)
								a.E[i, k] *= scale;
						}
					}
					anorm &= Max(anorm,(Abs(w.E[0, i]) + Abs(rv1.E[0, i])));
				}
				// accumulation of right-hand transformations
				for (i = n - 1; i >= 0; i--)
				{
					if (i < n - 1)
					{
						if (g != MultiFD.fZero)
						{
							for (j = L; j < n; j++)	// double division to avoid possible underflow
								v.E[j, i] &= (a.E[i, j] / a.E[i, L]) / g;
							for (j = L; j < n; j++)
							{
								for (s &= MultiFD.fZero, k = L; k < n; k++)
									s += a.E[i, k] * v.E[k, j];
								for (k = L; k < n; k++)
									v.E[k, j] += s * v.E[k, i];
							}
						}
						for (j = L; j < n; j++)
							v.E[i, j] &= v.E[j, i] &= MultiFD.fZero;
					}
					v.E[i, i] &= MultiFD.fOne;
					g &= rv1.E[0, i];
					L = i;
				}
				// accumulation of left-hand transformations
				for (i = Math.Min(m, n) - 1; i >= 0; i--)
				{
					L = i + 1;
					g &= w.E[0, i];
					for (j = L; j < n; j++)
						a.E[i, j] &= MultiFD.fZero;
					if (g != MultiFD.fZero)
					{
						g &= MultiFD.fOne / g;
						for (j = L; j < n; j++)
						{
							for (s &= MultiFD.fZero, k = L; k < m; k++)
								s += a.E[k, i] * a.E[k, j];
							f &= (s / a.E[i, i]) * g;
							for (k = i; k < m; k++)
								a.E[k, j] += f * a.E[k, i];
						}
						for (j = i; j < m; j++)
							a.E[j, i] *= g;
					}
					else
						for (j = i; j < m; j++)
							a.E[j, i] &= MultiFD.fZero;
					a.E[i, i] += MultiFD.fOne;
				}
				// diagonalization of the bidiagonal form
				for (k = n - 1; k >= 0; k--)				// loop over singular values
				{
					for (its = 1; its <= 300; its++)		// loop over allowed iterations
					{
						flag = 1;
						for (L = k; L >= 0; L--)		// test for splitting
						{								// note that rv1.E[0, 0] should always be zero
							nm = L - 1;
							if (L == 0 || (Abs(rv1.E[0, L]) + anorm) == anorm)
							{
								flag = 0;
								break;
							}
							if ((Abs(w.E[0, nm]) + anorm) == anorm)
								break;
						}
						if (flag != 0)
						{
							c &= MultiFD.fZero;			// cancellation of rv1.E[0, L], if L > 1
							s &= MultiFD.fOne;
							for (i = L; i < k + 1; i++)
							{
								f &= s * rv1.E[0, i];
								rv1.E[0, i] *= c;
								if ((Abs(f) + anorm) == anorm)
									break;
								g &= w.E[0, i];
								h &= Pythag(f, g);
								w.E[0, i] &= h;
								h &= MultiFD.fOne / h;
								c &= g * h;
								s &= -f * h;
								for (j = 0; j < m; j++)
								{
									y &= a.E[j, nm];
									z &= a.E[j, i];
									a.E[j, nm] &= y * c + z * s;
									a.E[j, i] &= z * c - y * s;
								}
							}
						}
						z &= w.E[0, k];
						if (L == k)						// convergence
						{
							if (z < MultiFD.fZero)		// singular value is made nonnegative
							{
								w.E[0, k] &= -z;
								for (j = 0; j < n; j++)
									v.E[j, k] &= -v.E[j, k];
							}
							break;
						}
						if (its >= 300)
						{
							MultiID.MuWriteErr("No convergence in 300 SVD iterations");
							return;
						}
						x &= w.E[0, L];				// shift from bottom 2-by-2 minor
						nm = k - 1;
						y &= w.E[0, nm];
						g &= rv1.E[0, nm];
						h &= rv1.E[0, k];
						f &= ((y - z) * (y + z) + (g - h) * (g + h)) / (MultiFD.fTwo * h * y);
						g &= Pythag(f, MultiFD.fOne);
						f &= ((x - z) * (x + z) + h*((y / (f + Sign(g, f))) - h)) / x;
						// next QR transform
						c &= s &= MultiFD.fOne;
						for (j = L; j <= nm; j++)
						{
							i = j + 1;
							g &= rv1.E[0, i];
							y &= w.E[0, i];
							h &= s * g;
							g &= c * g;
							z &= Pythag(f, h);
							rv1.E[0, j] &= z;			// can sets rv1.E[0, 0] > 0
							c &= f / z;
							s &= h / z;
							f &= x * c + g * s;
							g &= g * c - x * s;
							h &= y * s;
							y *= c;
							for (jj = 0; jj < n; jj++)
							{
								x &= v.E[jj, j];
								z &= v.E[jj, i];
								v.E[jj, j] &= x * c + z * s;
								v.E[jj, i] &= z * c - x * s;
							}
							z &= Pythag(f, h);
							w.E[0, j] &= z;			// rotation can be arbitrary if z == 0
							if (z != MultiFD.fZero)
							{
								z &= MultiFD.fOne / z;
								c &= f * z;
								s &= h *z ;
							}
							f &= c * g + s * y;
							x &= c * y - s * g;
							for (jj = 0; jj < m; jj++)
							{
								y &= a.E[jj, j];
								z &= a.E[jj, i];
								a.E[jj, j] &= y * c + z * s;
								a.E[jj, i] &= z * c - y * s;
							}
						}
						rv1.E[0, L] &= MultiFD.fZero;
						rv1.E[0, k] &= f;
						w.E[0, k] &= x;
					}
				}
				long next;
				long max = Int64.MinValue;
				for (i = 0; i < n; i++)
				{
					if (w.E[0, i] != MultiFD.fZero)
					{
						next = w.E[0, i].N + w.E[0, i].C;
						if (next > max)
							max = next;
					}
				}
				for (i = 0; i < n; i++)
				{
					next = w.E[0, i].N + w.E[0, i].C;
					if (max - next > Calc.fMC - MultiFD.fGuard)
						XPCalLst.SetToNM(MultiFD.fZero, w.E[0, i]);
				}
			}
			SortUWV(a, w, v);
		} // SVD

		public static void SVDN(MultiM x, int zI) // z = v part of SVD(x)
		{
			int m = x.Rs;
			int n = x.Cs;
			if (m == 0)
				MultiMD.ErrNotMatrix();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				if (m < n)
					m = n;
				MultiM u = new MultiM(Calc.fMC, m, n);
				MultiM w = new MultiM(Calc.fMC, 1, n);
				MultiM v = new MultiM(Calc.fMC, n, n);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < n; c++)
						MultiFD.SetToF(x.E[r, c], u.E[r, c]);
				SVD(u, w, v);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				int uI, wI, vI;
				Common.LFindAdd("MatU", out uI);
				Common.LFindAdd("MatW", out wI);
				Common.LFindAdd("MatV", out vI);
				Calc.SetToN(u, uI);
				Calc.SetToN(w, wI);
				Calc.SetToN(v, vI);
				Calc.SetToN(w, zI);
			}
		} // SVDN

		public static MultiF FacF(MultiF x) // return x!
		{
			int zI;
			Common.InitNAdd(1, "FacF-zI", out zI);
			XPCalLst.FacSN(x, zI, false, 0);
			MultiF z = new MultiF(Calc.fMC);
			XPCalLst.SetToNM(Common.nList[zI], z);
			Common.LDelete(zI);
			return z;
		} // FacF

		public static MultiF PowNF(MultiF b, MultiF p) // return z = b^p
		{
			MultiF z = new MultiF(Calc.fMC);
			XPCalLst.PowNF(b, p, z);
			return z;
		} // PowNF

		public static void Legen(MultiM x, MultiM z) // z = Legendre polynomial of Int(x.r)
		{
			if (x.Rs != 0)
				MultiMD.ErrNotScalar();
			else
			{
				MultiF nF = new MultiF();
				MultiF mF = new MultiF();
				MultiFD.IntF(x, nF);
				int n;
				MultiFD.Get1F(out n, nF);
				if (n < 0)
				{
					MultiID.MuWriteErr("Legen: argument must be >= 0");
					return;
				}
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiF f = new MultiF();
				MultiF t = new MultiF();
				MultiM zz = new MultiM(1, n + 1, 1);
				// z(n) = Fac(n+n) / (2^n * Fac(n)^2)
				XPCalLst.SetToNM(FacF(nF + nF) / (PowNF(MultiFD.fTwo, nF) * PowNF(FacF(nF), MultiFD.fTwo)), zz.E[0, 0]);
				for (int m = n - 2; m >= 0; m -= 2)
				{
					// f = (m+2)*(m+1) / ((n-m)*(n+m+1))
					MultiFD.SetTo1F(m, mF);
					XPCalLst.SetToNM(-zz.E[n - m - 2, 0] * (mF + MultiFD.fTwo) * (mF + MultiFD.fOne) /
						((nF - mF) * (nF + mF + MultiFD.fOne)), zz.E[n - m, 0]);
				}
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.SetToNM(zz, z);
			}
		} // Legen

		public static void LegenN(MultiM x, int zI) // z = Legendre polynomial of Int(x.r)
		{
			if (x.Rs != 0)
				MultiMD.ErrNotScalar();
			else
			{
				MultiM z = new MultiM();
				Legen(x, z);
				Calc.SetToN(z, zI);
			}
		} // LegenN

		public static void Cheby(MultiM x, MultiM z) // z = Chebyshev polynomial of Int(x.r)
		{
			if (x.Rs != 0)
				MultiMD.ErrNotScalar();
			else
			{
				MultiF nF = new MultiF();
				MultiFD.IntF(x, nF);
				nF.S = false;					// Cheby(-n) = Cheby(n)
				int n;
				MultiFD.Get1F(out n, nF);
				MultiM zz = new MultiM(1, n + 1, 1);
				if (n < 2)
				{
					zz.E[0, 0] &= MultiFD.fOne;	// Cheby(0) = {1}
					Calc.SetToNM(zz, z);		// Cheby(1) = {1; 0}
					return;
				}
				MultiM p0 = new MultiM(1, 1, 1);
				MultiM p1 = new MultiM(1, 2, 1);
				MultiM p2 = new MultiM(1, 2, 1);
				MultiM t = new MultiM();
				p0.E[0, 0] &= MultiFD.fOne;		// p1 = {1}
				p1.E[0, 0] &= MultiFD.fOne;		// p2 = {1; 0}
				p2.E[0, 0] &= MultiFD.fTwo;		// p2 = {2; 0}
				for (int i = 2; i <= n; i++)
				{
					PolMul(p2, p1, t);			// Cheby(n+1) = {2; 0} * Cheby(n)
					PolAdd(t, -p0, zz);			// - Cheby(n-1)
					if (i < n)
					{
						Calc.SetToNM(p1, p0);	// update Cheby(n-1)
						Calc.SetToNM(zz, p1);	// and Cheby(n)
					}
				}
				Calc.SetToNM(zz, z);
			}
		} // Cheby

		public static void ChebyN(MultiM x, int zI) // z = Chebyshev polynomial of Int(x.r)
		{
			if (x.Rs != 0)
				MultiMD.ErrNotScalar();
			else
			{
				MultiM z = new MultiM();
				Cheby(x, z);
				Calc.SetToN(z, zI);
			}
		} // ChebyN

		public static void DecompN(MultiM x, int zI) // z = LUdecomp(x)
		{
			if (x.Rs == 0)
				MultiMD.ErrNotMatrix();
			else if (x.Cs != x.Rs)
				MultiMD.ErrNotSquare();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int n = x.Rs;
				MultiM a = new MultiM(Calc.fMC, n, n);
				a &= x;
				int[] index;
				index = new int[n];
				MultiF d = new MultiF();
				bool sing;
				MultiMD.LUdecomp(a, index, d, out sing);	// decompose the matrix
				Calc.SetToN(a, zI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
			}
		} // DecompN

		public static void RREF(MultiM x, int e, MultiM z) // z = RREF of x less e right columns
			// RREF = Reduced Row Echelon Form
		{
			int m = x.Rs;
			int ne = x.Cs;
			int n = ne - e;
			int i = 0;
			int j = 0;
			int k;
			int c;
			int u;
			int maxInd;
			MultiF maxVal = new MultiF();
			MultiF val = new MultiF();
			MultiF temp = new MultiF();
			MultiM a = new MultiM();
			Calc.SetToNM(x, a);
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiF d = new MultiF(Calc.fMC - MultiFD.fGuard);
			MultiF auc = new MultiF(Calc.fMC - MultiFD.fGuard);
			while (i < m && j < n)
			{
				// Find pivot in column j, starting in row i
				XPCalLst.SetToNM(a.E[i, j], maxVal);
				maxInd = i;
				for (k = i+1; k < m; k++)
				{
					XPCalLst.SetToNM(a.E[k, j], val);
					if (Abs(val) > Abs(maxVal))
					{
						XPCalLst.SetToNM(val, maxVal);
						maxInd = k;
					}
				}
				if (maxVal != MultiFD.fZero)
				{
					// switch rows i and maxInd, but remain the values of i and maxInd
					// Now a[i, j] will contain the same value as maxVal
					for (c = 0; c < ne; c++)
					{
						if (c != j)
							XPCalLst.SetToNM(a.E[maxInd, c], temp);
						XPCalLst.SetToNM(a.E[i, c], a.E[maxInd, c]);
						// divide row i by maxVal
						if (c != j)
							XPCalLst.SetToNM(temp / maxVal, a.E[i, c]);
					}
					// Now A[i, j] will have the value 1
					XPCalLst.SetToNM(MultiFD.fOne, a.E[i, j]);
					for (u = 0; u < m; u++)
					{
						if (u != i)
						{
							// subtract a[u, j] * row i from row u
							for (c = 0; c < ne; c++)
								if (c != j)
								{
									auc &= a.E[u, c];
									XPCalLst.SetToNM(a.E[u, c] - a.E[u, j] * a.E[i, c], a.E[u, c]);
									d &= auc - a.E[u, c];
									if (d == auc)
										XPCalLst.SetToNM(MultiFD.fZero, a.E[u, c]);
								}
							// Now A[u, j] will be 0, since
							// A[u, j] - A[u, j] * A[i, j] = A[u, j] - 1 * A[u, j] = 0
							XPCalLst.SetToNM(MultiFD.fZero, a.E[u, j]);
						}
					}
					i++;
				}
				j++;
			}
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.SetToNM(a, z);
		} // RREF

		public static void RREF(MultiM x, MultiM z) // z = RREF of x
		// RREF = Reduced Row Echelon Form
		{
			RREF(x, 0, z);
		} // RREF

		public static void RREFN(MultiM x, int zI) // z = RREF of x
		// RREF = Reduced Row Echelon Form
		{
			if (x.Rs == 0)
				MultiMD.ErrNotMatrix();
			else
			{
				MultiM z = new MultiM();
				RREF(x, z);
				Calc.SetToN(z, zI);
			}
		} // RREFN

		public static void RREFa(MultiM x, MultiM y, MultiM z) // z = RREF of [x y]
		// RREF = Reduced Row Echelon Form
		{
			int tI;
			Common.InitNAdd(1, "RREFa-tI", out tI);
			ConcatN(x, y, tI);
			MultiM zz = new MultiM();
			RREF(Common.nList[tI], y.Cs, zz);
			if (x.Rs != x.Cs || zz.E[x.Rs - 1, x.Cs - 1] == MultiFD.fZero)
				MultiID.WriteDLn("Matrix is singular");
			MatrixN(y.Rs, y.Cs, tI);
			for (int r = 0; r < y.Rs; r++)
				for (int c = 0; c < y.Cs; c++)
					XPCalLst.SetToNM(zz.E[r, c + x.Cs], Common.nList[tI].E[r, c]);
			Calc.SetToNM(Common.nList[tI], z);
			Common.LDelete(tI);
		} // RREFa

		public static void RREFaN(MultiM x, MultiM y, int zI) // z = RREF of [x y]
		// RREF = Reduced Row Echelon Form
		{
			if (x.Rs == 0 || y.Rs == 0)
				MultiMD.ErrNotMatrix();
			else if (x.Rs != y.Rs)
				MultiMD.ErrIncompatible();
			else
			{
				MultiM z = new MultiM();
				RREFa(x, y, z);
				Calc.SetToN(z, zI);
			}
		} // RREFaN

		public static void InvN(MultiM x, int zI) // z = 1 / x
		// this code is adapted from Numerical Recipes in C
		{
			if (x.Rs == 0)
				XPCalLst.InvN(x, zI);
			else if (x.Cs != x.Rs)
				MultiMD.ErrNotSquare();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiM z = new MultiM(Calc.fMC, x.Rs, x.Rs);
				MultiMD.InvM(x, z);
				Calc.SetToN(z, zI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
			}
		} // InvN

		public static void RInvN(int zI) // z = 1 / z
		{
			InvN(Common.nList[zI], zI);
		}

		public static void SetFMBN(int zI) // set Modulo base FMB on Common.nList = value of zI
		{
			int fMBI;
			Common.LFindAdd("FMB", out fMBI);
			if (fMBI == zI) return;
			Calc.SetToII(zI, fMBI);
		}

		public static void ModFMBN(MultiM x, int zI) // z = x % FMB
		{
			int fMBI;
			if (Common.FMBIok(zI, out fMBI) && (!Common.ZTestN(fMBI) || Common.nList[fMBI].Rs != 0))
				RModN(Common.nList[fMBI], zI);
		}

		public static void RModFMBN(int zI) // z = z % FMB
		{
			ModFMBN(Common.nList[zI], zI);
		}

		public static void SinN(MultiM x, int zI)	// z = Sin(x)
		{
			if (x.Rs == 0)
				XPCalLst.SinN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "SinN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.SinN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // SinN

		public static void RSinN(int zI)	// z = Sin(z)
		{
			SinN(Common.nList[zI], zI);
		}

		public static void SignN(MultiM x, int zI)	// z = Sin(x)
		{
			if (x.Rs == 0)
				XPCalLst.SignN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "SignN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.SignN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // SignN

		public static void RSignN(int zI)	// z = Sin(z)
		{
			SignN(Common.nList[zI], zI);
		}

		public static void CosN(MultiM x, int zI)	// z = Cos(x)
		// Cos(x) = Sin(x + pi/2)
		{
			if (x.Rs == 0)
				XPCalLst.CosN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "CosN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.CosN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // CosN

		public static void RCosN(int zI) // z = Cos(z)
		{
			CosN(Common.nList[zI], zI);
		}

		public static void TanN(MultiM x, int zI)	// z = Tan(x)
		// Tan(x) = Sin(x) / SqRt(1 - Sq(Sin(x))
		// and change sign of Tan(x) if in 2nd or 3rd quadrant
		{
			bool forCot = false;
			TanN(x, zI, ref forCot);
		}

		public static void TanN(MultiM x, int zI, ref bool forCot)	// z = Tan(x)
		// Tan(x) = Sin(x) / SqRt(1 - Sq(Sin(x))
		// and change sign of Tan(x) if in 2nd or 3rd quadrant
		{
			if (x.Rs == 0)
				XPCalLst.TanN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "TanN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.TanN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // TanN

		public static void RTanN(int zI) // z = Tan(z)
		{
			TanN(Common.nList[zI], zI);
		}

		public static void ATanN(MultiM x, int zI)	// z = ATan(x)
		{
			if (x.Rs == 0)
				XPCalLst.ATanN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ATanN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.ATanN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ATanN

		public static void RATanN(int zI) // z = ATan(z)
		{
			ATanN(Common.nList[zI], zI);
		}

		public static void ATan2N(MultiM y, MultiM x, int zI)	// z = ATan(y / x)
		// Compute Arctan(y / x) from -pi to pi
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.ATan2N(y, x, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "ATan2N-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.ATan2N(y.E[r, c], x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ATan2N

		public static void AbsN(MultiM x, int zI) // z = |x|
		{
			Calc.SetToN(x, zI);
			if (x.Rs == 0)
				Common.nList[zI].S = false;
			else
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
						Common.nList[zI].E[r, c].S = false;
		}

		public static void MagN(MultiM x, MultiM y, int zI)	// z = SqRt(Sq(x) + Sq(y))
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.MagN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "MagN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.MagN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // MagN

		public static void MeanCN(MultiM x, MultiM y, int zI)	// z = common mean of x and y
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.MeanCN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "MeanCN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.MeanCN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} //MeanCN

		public static void MaxN(MultiM x, MultiM y, int zI)	// z = max of x and y
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.MaxN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "MaxN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.MaxN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} //MaxN

		public static void MinN(MultiM x, MultiM y, int zI)	// z = min of x and y
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.MinN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "MinN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.MinN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} //MinN

		public static void ModN(MultiM x, MultiM y, int zI)	// z = Mod(x, y))
		// same as FORTRAN definition: z = Mod(x, y) = x - (Floor(x/y) * y)
		// sign(z) = sign(y) or z = 0
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.ModN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "ModN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.ModN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ModN int

		public static void RModN(MultiM x, int zI) // z = Mod(z, x))
		{
			ModN(Common.nList[zI], x, zI);
		}

		public static void GamPN(MultiM x, MultiM y, int zI) // z = GamP(x, y)
		// GamP(a, x) is the lower regularized incomplete gamma function with parameter a and argument x
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.GamPN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "GamPN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.GamPN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // GamPN

		public static void GamQN(MultiM x, MultiM y, int zI) // z = GamQ(x, y)
		// GamQ(a, x) is the upper regularized incomplete gamma function with parameter a and argument x
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.GamQN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "GamQN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.GamQN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // GamQN

		public static void GamLN(MultiM x, MultiM y, int zI) // z = GamL(x, y)
		// GamL(a, x) is the lower incomplete gamma function with parameter a and argument x
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.GamLN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "GamLN", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.GamLN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // GamLN

		public static void GamUN(MultiM x, MultiM y, int zI) // z = GamU(x, y)
		// GamU(a, x) is the upper incomplete gamma function with parameter a and argument x
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.GamUN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "GamUN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.GamUN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // GamUN

		public static void BetaCN(MultiM x, MultiM y, int zI)
		// z = the complete beta function B(x, y) = gamma(x) * gamma(y) / gamma(x+y)
		// also equal to 1 / (x * Bino(x+y-1, y-1)
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.BetaCN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "BetaCN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.BetaCN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // BetaCN

		public static void BinoGN(MultiM x, MultiM y, int zI)
		// z = Binomial coefficient (x, y) using the gamma function
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.BinoGN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "BinoGN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.BinoGN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // BinoGN

		public static void GCD(MultiM a, MultiM b, MultiM z) // z = GCD(a, b))
		{
			if (a.Rs == 0 && b.Rs == 0)
				XPCalLst.GCD(a, b, z);
			else if (a.Rs != b.Rs || a.Cs != b.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "GCD-tI", out tI);
				if (z.Rs != a.Rs || z.Cs != a.Cs)
					Matrix(a.Rs, a.Cs, z);
				for (int r = 0; r < a.Rs; r++)
					for (int c = 0; c < a.Cs; c++)
					{
						XPCalLst.GCDN(a.E[r, c], b.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], z.E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // GCD

		public static void GCDN(MultiM a, MultiM b, int zI) // z = GCD(a, b))
		{
			MultiM zz = new MultiM();
			GCD(a, b, zz);
			Calc.SetToN(zz, zI);
		} // GCDN

		public static void LCM(MultiM a, MultiM b, MultiM z) // z = LCM(a, b))
		// LCM(a, b) = a * b / GCD(a, b)
		{
			if (a.Rs == 0 && b.Rs == 0)
				XPCalLst.LCM(a, b, z);
			else if (a.Rs != b.Rs || a.Cs != b.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "LCM-tI", out tI);
				if (z.Rs != a.Rs || z.Cs != a.Cs)
					Matrix(a.Rs, a.Cs, z);
				for (int r = 0; r < a.Rs; r++)
					for (int c = 0; c < a.Cs; c++)
					{
						XPCalLst.LCMN(a.E[r, c], b.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], z.E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // LCM

		public static void LCMN(MultiM a, MultiM b, int zI) // z = LCM(a, b))
		// LCM(a, b) = a * b / GCD(a, b)
		{
			MultiM zz = new MultiM();
			LCM(a, b, zz);
			Calc.SetToN(zz, zI);
		} // LCMN

		public static void ASinN(MultiM x, int zI)	// z = ASin(x)
		// ASin(S) = ATan2(S, SqRt(1 - Sq(S)), but if Sq(S) > 1 then Error
		{
			if (x.Rs == 0)
				XPCalLst.ASinN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ASinN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.ASinN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ASinN

		public static void RASinN(int zI) // z = ASin(z)
		{
			ASinN(Common.nList[zI], zI);
		}

		public static void ACosN(MultiM x, int zI)	// z = ACos(x)
		// ACos(C) = ATan2(SqRt(1 - Sq(C), C), but if Sq(C) > 1 then Error
		{
			if (x.Rs == 0)
				XPCalLst.ACosN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ACosN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.ACosN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ACosN

		public static void RACosN(int zI) // z = ACos(z)
		{
			ACosN(Common.nList[zI], zI);
		}

		public static void ExpLN(MultiM x, bool bL1, int zI)
		// z = (e^x) - 1, accurate for small x, (- 1 only if bL1 is true)
		// n = Math.Floor(x / Ln(10))
		// g = Frac(x / Ln(10)) * Ln(10)
		// answer = Exp(g) * 10^n
		{
			if (x.Rs == 0)
				XPCalLst.ExpLN(x, bL1, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ExpLN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.ExpLN(x.E[r, c], bL1, tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ExpLN

		public static void RExpLN(int zI) // z = (e^z) - 1, accurate for small x
		{
			ExpLN(Common.nList[zI], true, zI);
		}

		public static void ExpN(MultiM x, int zI)	// z = e^x
		{
			ExpLN(x, false, zI);
		}

		public static void RExpN(int zI) // z = e^z
		{
			ExpN(Common.nList[zI], zI);
		}

		public static void LnLN(MultiM y, bool bL1, int zI)
		// z = Ln(y + 1), accurate for small y
		// x = y + 1 = Integer * MultiID.baseM^C
		// Multiply by 10 until MSD of i >= MultiID.baseM / 10, save count
		// change C of x to -N so x < 1, save original C+N
		// answer = Ln(x) + Ln(10) * NT, NT = MultiID.muDMax * C+N - count
		{
			if (y.Rs == 0)
				XPCalLst.LnLN(y, bL1, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "LnLN-tI", out tI);
				if (Common.nList[zI].Rs != y.Rs || Common.nList[zI].Cs != y.Cs)
					MatrixN(y.Rs, y.Cs, zI);
				for (int r = 0; r < y.Rs; r++)
					for (int c = 0; c < y.Cs; c++)
					{
						XPCalLst.LnLN(y.E[r, c], bL1, tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // LnLN

		public static void RLnLN(int zI) // z = Ln(z + 1), accurate for small z
		{
			LnLN(Common.nList[zI], true, zI);
		}

		public static void LnN(MultiM x, int zI)	// z = Ln(x)
		{
			LnLN(x, false, zI);
		}

		public static void RLnN(int zI) // z = Ln(z)
		{
			LnN(Common.nList[zI], zI);
		}

		public static void LogN(MultiM x, int zI)	// z = Log(x)
		// Log(x) = Ln(x) / Ln(10), Log base 10 of x
		{
			if (x.Rs == 0)
				XPCalLst.LogN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "LogN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.LogN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		}

		public static void RLogN(int zI) // z = Log(z)
		{
			LogN(Common.nList[zI], zI);
		}

		public static void SinHN(MultiM x, int zI)	// z = SinH(x)
		// for (x >= 0.1 SinH(x) = (y - 1/y) / 2, where y = Exp(x)
		// for (x <  0.1 SinH(x) = y / (2 * SqRt(y + 1)), where y = Exp(2 * x) - 1
		// SinH(-x) = -SinH(x)
		{
			if (x.Rs == 0)
				XPCalLst.SinHN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "SinHN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.SinHN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // SinHN

		public static void RSinHN(int zI) // z = SinH(z)
		{
			SinHN(Common.nList[zI], zI);
		}

		public static void RiN(MultiM x, int zI)	// z = Riemann prime counting function of x
		// Ri(x) = 1 + sum{n = 1, ...} ( Ln(x)^n / (n*n!*Zeta(n+1)) )
		{
			if (x.Rs == 0)
				XPCalLst.RiN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "RiN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.RiN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RiN

		public static void LiN(MultiM x, int zI)	// z = Logarithmic integral of x
		// Li(x) = Ei(Ln(x))
		{
			if (x.Rs == 0)
				XPCalLst.LiN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "LiN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.LiN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // LiN

		public static void EiN(MultiM x, int zI)	// z = entire exponential integral of x
		// Ei(x) = -Ei(-x)
		{
			if (x.Rs == 0)
				XPCalLst.EiN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "EiN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.EiN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // EiN

		public static void EiAN(MultiM x, int zI)	// z = entire exponential integral of x
		// asymptotic expansion, only use this for x << -1 (large negative x)
		{
			if (x.Rs == 0)
				XPCalLst.EiAN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "EiAN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.EiAN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // EiAN

		public static void E1N(MultiM x, int zI)	// z = exponential integral 1 of x
		// E1(x) = integral{x to inf}(e^-t / t)dt = -Ei(-x)
		{
			if (x.Rs == 0)
				XPCalLst.E1N(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "E1N-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.E1N(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // E1N

		public static void DilogN(MultiM x, int zI)	// z = Dilog(x)
		// Dilog(x) = Polylog(2, x) = x * Lerch(x, 2, 1)
		{
			if (x.Rs == 0)
				PolylogN(MultiMD.mTwo, x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "DilogN-tI", out tI);
				MatrixN(x.Rs, x.Cs, tI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.SetToNM(MultiFD.fTwo, Common.nList[tI].E[r, c]);
					}
				PolylogN(Common.nList[tI], x, zI);
				Common.LDelete(tI);
			}
		} // DilogN

		public static void PolylogN(MultiM x, MultiM y, int zI)	// z = Polylog(x, y)
		// Polylog(s, x) = x * Lerch(x, s, 1)
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.PolylogN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "PolylogN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.PolylogN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // PolylogN

		public static void ZetaHN(MultiM s, MultiM a, int zI)	// z = ZetaH(s, a) Hurwitz Zeta Function
		// ZetaH(s, a) = Lerch(1, s, a)
		{
			if ((s.Rs == 0 && a.Rs == 0) || s.Rs != a.Rs || s.Cs != a.Cs)
				Lerch(MultiMD.mOne, s, a, false, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ZetaHN-tI", out tI);
				MatrixN(s.Rs, s.Cs, tI);
				for (int r = 0; r < s.Rs; r++)
					for (int c = 0; c < s.Cs; c++)
					{
						XPCalLst.SetToNM(MultiFD.fOne, Common.nList[tI].E[r, c]);
					}
				Lerch(Common.nList[tI], s, a, false, zI);
				Common.LDelete(tI);
			}
		} // ZetaHN

		public static void CosHN(MultiM x, int zI)	// z = CosH(x)
		// CosH(x) = (y + 1/y) / 2, where y = Exp(|x|)
		{
			if (x.Rs == 0)
				XPCalLst.CosHN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "CosHN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.CosHN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // CosHN

		public static void RCosHN(int zI) // z = CosH(z)
		{
			CosHN(Common.nList[zI], zI);
		}

		public static void TanHN(MultiM x, int zI)	// z = TanH(x)
		// TanH(x) = y / (y + 2), where y = Exp(2 * x) - 1
		// TanH(-x) = -TanH(x)
		{
			if (x.Rs == 0)
				XPCalLst.TanHN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "TanHN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.TanHN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // TanHN

		public static void RTanHN(int zI) // z = TanH(z)
		{
			TanHN(Common.nList[zI], zI);
		}

		public static void ASinHN(MultiM x, int zI)	// z = ASinH(x)
		// for (x >= 0.1 ASinH(x) = Ln(x + SqRt(1 + Sq(x)))
		// for (x <  0.1 ASinH(x) = LnL(x + Sq(x) / SqRt(1 + Sq(x)))
		// ASinH(-x) = -ASinH(x)
		{
			if (x.Rs == 0)
				XPCalLst.ASinHN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ASinHN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.ASinHN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ASinHN

		public static void RASinHN(int zI) // z = ASinH(z)
		{
			ASinHN(Common.nList[zI], zI);
		}

		public static void ACosHN(MultiM x, int zI)	// z = ACosH(x)
		// ACosH(x) = Ln(x + SqRt(Sq(x) - 1))
		// error if x < 1
		{
			if (x.Rs == 0)
				XPCalLst.ACosHN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ACosHN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.ACosHN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ACosHN

		public static void RACosHN(int zI) // z = ACosH(z)
		{
			ACosHN(Common.nList[zI], zI);
		}

		public static void RACotN(int zI) // z = ACot(z)
		// ACot(z) = ATan(1/z), error if z = 0
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XPCalLst.RACotN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RACotN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						XPCalLst.SetToNM(Common.nList[zI].E[r, c], (MultiF)Common.nList[tI]);
						XPCalLst.RACotN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RACotN

		public static void RACotHN(int zI) // z = ACotH(z)
		// ACotH(z) = ATanH(1/z), error if |z| <= 1
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XPCalLst.RACotHN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RACotHN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						XPCalLst.SetToNM(Common.nList[zI].E[r, c], (MultiF)Common.nList[tI]);
						XPCalLst.RACotHN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RACotHN

		public static void RACscN(int zI) // z = ACsc(z)
		// ACsc(z) = ASin(1/z), error if z = 0
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XPCalLst.RACscN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RACscN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						XPCalLst.SetToNM(Common.nList[zI].E[r, c], (MultiF)Common.nList[tI]);
						XPCalLst.RACscN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RACscN

		public static void RACscHN(int zI) // z = ACscH(z)
		// ACscH(z) = ASinH(1/z), error if z = 0
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XPCalLst.RACscHN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RACscHN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						XPCalLst.SetToNM(Common.nList[zI].E[r, c], (MultiF)Common.nList[tI]);
						XPCalLst.RACscHN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RACscHN

		public static void RASecN(int zI) // z = ASec(z)
		// ASec(z) = ACos(1/z), error if z = 0
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XPCalLst.RASecN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RASecN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						XPCalLst.SetToNM(Common.nList[zI].E[r, c], (MultiF)Common.nList[tI]);
						XPCalLst.RASecN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RASecN

		public static void RASecHN(int zI) // z = ASecH(z)
		// ASecH(z) = ACosH(1/z), error if z <= 0 or z > 1
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XPCalLst.RASecHN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RASecHN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						XPCalLst.SetToNM(Common.nList[zI].E[r, c], (MultiF)Common.nList[tI]);
						XPCalLst.RASecHN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RASecHN

		public static void RCotN(int zI) // z = Cot(z)
		// Cot(z) = 1/Tan(z), error if Tan(z) = 0
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XPCalLst.RCotN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RCotN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						XPCalLst.SetToNM(Common.nList[zI].E[r, c], (MultiF)Common.nList[tI]);
						XPCalLst.RCotN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RCotN

		public static void RCotHN(int zI) // z = CotH(z)
		// CotH(z) = 1/TanH(z), error if TanH(z) = 0
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XPCalLst.RCotHN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RCotHN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						XPCalLst.SetToNM(Common.nList[zI].E[r, c], (MultiF)Common.nList[tI]);
						XPCalLst.RCotHN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RCotHN

		public static void RCscN(int zI) // z = Csc(z)
		// Csc(z) = 1/Sin(z), error if Sin(z) = 0
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XPCalLst.RCscN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RCscN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						XPCalLst.SetToNM(Common.nList[zI].E[r, c], (MultiF)Common.nList[tI]);
						XPCalLst.RCscN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RCscN

		public static void RCscHN(int zI) // z = CscH(z)
		// CscH(z) = 1/SinH(z), error if SinH(z) = 0
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XPCalLst.RCscHN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RCscHN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						XPCalLst.SetToNM(Common.nList[zI].E[r, c], (MultiF)Common.nList[tI]);
						XPCalLst.RCscHN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RCscHN

		public static void RSecN(int zI) // z = Sec(z)
		// Sec(z) = 1/Cos(z), error if Cos(z) = 0
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XPCalLst.RSecN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RSecN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						XPCalLst.SetToNM(Common.nList[zI].E[r, c], (MultiF)Common.nList[tI]);
						XPCalLst.RSecN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RSecN

		public static void RSecHN(int zI) // z = SecH(z)
		// SecH(z) = 1/CosH(z)
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XPCalLst.RSecHN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RSecHN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						XPCalLst.SetToNM(Common.nList[zI].E[r, c], (MultiF)Common.nList[tI]);
						XPCalLst.RSecHN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RSecHN

		public static void ATanHN(MultiM x, int zI)	// z = ATanH(x)
		// ATanH(x) = LnL(2 * x / (1 - x))
		// ATanH(-x) = -ATanH(x)
		// error if |x| >= 1
		{
			if (x.Rs == 0)
				XPCalLst.ATanHN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ATanHN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.ATanHN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ATanHN

		public static void RATanHN(int zI) // z = ATanH(z)
		{
			ATanHN(Common.nList[zI], zI);
		}

		public static void CeilingN(MultiM x, int zI)	// z = smallest integer >= z
		{
			if (x.Rs == 0)
				XPCalLst.CeilingN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "CeilingN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.CeilingN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // CeilingN

		public static void RCeilingN(int zI) // z = smallest integer >= z
		{
			CeilingN(Common.nList[zI], zI);
		}

		public static void FloorN(MultiM x, int zI)	// z = largest integer <= z
		{
			if (x.Rs == 0)
				XPCalLst.FloorN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "FloorN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.FloorN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // FloorN

		public static void RFloorN(int zI) // z = largest integer <= z
		{
			FloorN(Common.nList[zI], zI);
		}

		public static void RoundN(MultiM x, int zI)	// z = integer nearest to x
		{
			if (x.Rs == 0)
				XPCalLst.RoundN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "RoundN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.RoundN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RoundN

		public static void RRoundN(int zI) // z = integer nearest to z
		{
			RoundN(Common.nList[zI], zI);
		}

		public static void IntN(MultiM x, int zI)	// z = Integer part of x
		{
			if (x.Rs == 0)
				XPCalLst.IntN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "IntN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.IntN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // IntN

		public static void RIntN(int zI) // z = Integer part of z
		{
			IntN(Common.nList[zI], zI);
		}

		public static void FracN(MultiM x, int zI)	// z = Fractional part of x
		{
			if (x.Rs == 0)
				XPCalLst.FracN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "FracN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.FracN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // FracN

		public static void RFracN(int zI) // z = Fractional part of z
		{
			FracN(Common.nList[zI], zI);
		}

		public static void LopN(MultiM x, int zI) // z = x with one less super digit
		{
			if (x.Rs == 0)
				XPCalLst.LopN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "LopN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.LopN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // LopN

		public static void RLopN(int zI) // z = z with one less least significant "digit"
		{
			MultiMD.RLopM(Common.nList[zI]);
			Calc.NormN(zI);
		}

		public static void ToDegN(MultiM x, int zI) // z = x converted Rad to Degrees
		{
			if (x.Rs == 0)
				XPCalLst.ToDegN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ToDegN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.ToDegN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ToDegN

		public static void RToDegN(int zI) // z = z converted Radians to Degrees
		{
			ToDegN(Common.nList[zI], zI);
		}

		public static void ToRadN(MultiM x, int zI)	// z = x converted Deg to Radians
		{
			if (x.Rs == 0)
				XPCalLst.ToRadN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ToRadN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.ToRadN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ToRadN

		public static void RToRadN(int zI) // z = z converted Deg to Radians
		{
			ToRadN(Common.nList[zI], zI);
		}

		public static void CalSFTn(int zI) // set number of digits in truncate in display
		{
			if (Common.nList[zI].Rs != 0)
				MultiMD.ErrNotScalar();
			else
				XPCalLst.CalSFTn(zI);
		}

		public static void NormM(MultiM x, MultiM m, MultiM s, int zI)
		// z = Cumulative distribution function (cdf) of normal distribution
		// m = mean, s = standard deviation
		// x, m, s may have the same address in memory as z
		// x, m, s are not changed unless it shares memory with z
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0 && m.Rs == 0 && s.Rs == 0)
				XPCalLst.NormC(x, m, s, zI);
			else if (x.Rs != m.Rs || x.Rs != s.Rs || x.Cs != m.Cs || x.Cs != s.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "NormC-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.NormC(x.E[r, c], m.E[r, c], s.E[r, c],tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // NormC

		public static void NormS(int zI)
		// z = Cumulative distribution function (cdf) of standard normal distribution
		// mean = 0, standard deviation = 1
		// sets muErr true if overflow or error
		{
			if (Common.nList[zI].Rs == 0)
				NormM(Common.nList[zI], MultiMD.mZero, MultiMD.mOne, zI);
			else
			{
				int t0I;
				int t1I;
				Common.InitNAdd(1, "NormS-t0I", out t0I);
				Common.InitNAdd(1, "NormS-t1I", out t1I);
				MatrixN(Common.nList[zI].Rs, Common.nList[zI].Cs, t0I);
				MatrixN(Common.nList[zI].Rs, Common.nList[zI].Cs, t1I);
				for (int r = 0; r < Common.nList[zI].Rs; r++)
					for (int c = 0; c < Common.nList[zI].Cs; c++)
					{
						XPCalLst.SetToNM(MultiFD.fZero, Common.nList[t0I].E[r, c]);
						XPCalLst.SetToNM(MultiFD.fOne, Common.nList[t1I].E[r, c]);
					}
				NormM(Common.nList[zI], Common.nList[t0I], Common.nList[t1I], zI);
				Common.LDelete(t1I);
				Common.LDelete(t0I);
			}
		} // NormS

		public static void QToMN(MultiM x, int zI)	// z = rotation matrix from rotation quaternion
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiM z = new MultiM(Calc.fMC, 3, 3);
			MultiMD.QToM(x, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // QToMN

		public static void MToQN(MultiM x, int zI)	// z = rotation quaternion from rotation matrix
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiM z = new MultiM(Calc.fMC, 4, 1);
			MultiMD.MToQ(x, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // MToQN

		public static void MToRN(MultiM x, int zI)	// z = rotation vector from rotation matrix
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiM z = new MultiM(Calc.fMC, 4, 1);
			MultiMD.MToR(x, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // MToRN

		public static void RToMN(MultiM x, int zI)	// z = rotation matrix from rotation vector
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiM z = new MultiM(Calc.fMC, 3, 3);
			MultiMD.RToM(x, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // RToMN

		public static void QToRN(MultiM x, int zI)	// z = rotation vector from rotation quaternion
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiM z = new MultiM(Calc.fMC, 4, 1);
			MultiMD.QToR(x, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // QToRN

		public static void RToQN(MultiM x, int zI)	// z = rotation quaternion from rotation vector
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiM z = new MultiM(Calc.fMC, 4, 1);
			MultiMD.RToQ(x, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // RToQN

		public static void PrinQN(MultiM x, int zI)	// z = principal quaternion of x and -x
		{
			MultiM z = new MultiM(Calc.fMC, 4, 1);
			MultiMD.PrinQ(x, z);
			Calc.SetToN(z, zI);
			Calc.NormN(zI);
		} // PrinQN

		public static void RotateQN(MultiM x, MultiM y, int zI)	// z = quaternion rotation x*y
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiM z = new MultiM(Calc.fMC, 3, 1);
			MultiMD.RotateQ(x, y, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // RotateQN

		public static void RotateMN(MultiM x, MultiM y, int zI)	// z = 3x3 matrix rotation x*y
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiM z = new MultiM(Calc.fMC, 3, 1);
			MultiMD.RotateM(x, y, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // RotateMN

		public static void EqQN(MultiM x, int zI)	// z = quaternion z normalized
		{
			MultiM z = new MultiM(Calc.fMC, 4, 1);
			MultiMD.EqQ(x, z);
			Calc.SetToN(z, zI);
			Calc.NormN(zI);
		} // EqQN

		public static void InvQN(MultiM x, int zI)	// z = quaternion z normalized
		{
			MultiM z = new MultiM(Calc.fMC, 4, 1);
			MultiMD.InvQ(x, z);
			Calc.SetToN(z, zI);
			Calc.NormN(zI);
		} // InvQN

		public static void NormQN(MultiM x, int zI)	// z = quaternion z normalized
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiM z = new MultiM(Calc.fMC, 4, 1);
			MultiMD.NormQ(x, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // NormQN

		public static void NormRN(MultiM x, int zI)	// z = rotation vector z normalized
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiM z = new MultiM(Calc.fMC, 4, 1);
			MultiMD.NormR(x, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // NormRN

		public static void NormVN(MultiM x, int zI)
		// z = quaternion z normalized
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiM z = new MultiM(Calc.fMC);
			MultiMD.NormV(x, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // NormVN

		public static void VanderN(MultiM x, int zI) // z = Vandermonde's matrix from vector x
		// zij = xi^(j-1)
		{

			if (x.Rs != 1 && x.Cs != 1)
				MultiMD.ErrNotVector();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int n = Math.Max(x.Rs, x.Cs);
				MultiM z = new MultiM(Calc.fMC, n, n);
				MultiF jC = new MultiF(1);
				if (x.Rs == 1)
					for (int i = 0; i < n; i++)
						for (int j = 0; j < n; j++)
						{
							jC &= j;
							XPCalLst.PowNF(x.E[0, i], jC, z.E[i, j]);
						}
				else
					for (int i = 0; i < n; i++)
						for (int j = 0; j < n; j++)
						{
							jC &= j;
							XPCalLst.PowNF(x.E[i, 0], jC, z.E[i, j]);
						}
				Calc.SetToN(z, zI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
			}
		} // VanderVN

		public static void QuadRoot(MultiM a, MultiM b, MultiM c, int zI)
		// z = a root of quadratic equation 0 = a*x^2 + b*x + c, roots also stored in X1 and X2
		// a, b, c may have the same address in memory as z
		// a, b, c are not changed unless it shares memory with z
		// sets muErr true if overflow or error
		{
			if (a.Rs == 0 && b.Rs == 0 && c.Rs == 0)
				XPCalLst.QuadRoot(a, b, c, zI);
			else if (a.Rs != b.Rs || a.Cs != b.Cs || a.Rs != c.Rs || a.Cs != c.Cs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI;
				int x1MI;
				int x2MI;
				Common.InitNAdd(1, "QuadRoot-tI", out tI);
				Common.InitNAdd(1, "QuadRoot-x1MI", out x1MI);
				Common.InitNAdd(1, "QuadRoot-x2MI", out x2MI);
				MatrixN(a.Rs, a.Cs, x1MI);
				MatrixN(a.Rs, a.Cs, x2MI);
				if (Common.nList[zI].Rs != a.Rs || Common.nList[zI].Cs != a.Cs)
					MatrixN(a.Rs, a.Cs, zI);
				int x1I;
				int x2I;
				Common.LFindAdd("X1", out x1I);
				Common.LFindAdd("X2", out x2I);
				for (int r = 0; r < a.Rs; r++)
					for (int co = 0; co < a.Cs; co++)
					{
						XPCalLst.QuadRoot(a.E[r, co], b.E[r, co], c.E[r, co], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(x2MI);
							Common.LDelete(x1MI);
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[x1I], Common.nList[x1MI].E[r, co]);
						XPCalLst.SetToNM((MultiF)Common.nList[x2I], Common.nList[x2MI].E[r, co]);
					}
				Calc.SetToII(x1MI, x1I);
				Calc.SetToII(x2MI, x2I);
				Common.LDelete(x2MI);
				Common.LDelete(x1MI);
				Common.LDelete(tI);
			}
		} // QuadRoot

		public static void ChSMN(int zI) // change sign, n = - n
		{
			MultiMD.ChSM(Common.nList[zI]);
		}

		public static bool ZTestN(int zI) // true if z is Zero
		{
			return MultiMD.ZTestM(Common.nList[zI]);
		}

		public static void Matrix(int rs, int cs, MultiM z)
			// z = zero matrix with rs rows and cs columns
		{
			MultiM zz;
			MultiMD.Matrix(0, rs, cs, out zz);
			Calc.SetToNM(zz, z);
		} // Matrix

		public static void MatrixN(MultiM rsF, MultiM csF, int zI)
		// z = zero matrix with rs rows and cs columns
		{
			if (rsF.Rs != 0 || csF.Rs != 0)
				MultiMD.ErrNotScalar();
			else
			{
				int rs;
				int cs;
				MultiFD.Get1F(out rs, rsF);
				MultiFD.Get1F(out cs, csF);
				MatrixN(rs, cs, zI);
			}
		} // MatrixN

		public static void MatrixN(int rs, int cs, int zI)
		// z = zero matrix with rs rows and cs columns
		{
			MultiM z;
			MultiMD.Matrix(0, rs, cs, out z);
			Calc.SetToN(z, zI);
		} // MatrixN

		public static void CharPoly(MultiM x, MultiM z)
		// z = Characteristic polynomial of square matrix x
		{
			int n = x.Rs;
			MultiM zz;
			if (x.Cs != n)
				MultiMD.ErrNotSquare();
			else if (n == 0)					// if a scalar
			{

				zz = new MultiM(1, 2, 1);
				zz.E[0, 0] &= MultiFD.fOne;
				XPCalLst.SetToNM(-x, zz.E[1, 0]);
				Calc.SetToNM(zz, z);
			}
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int i, j, k, m;
				zz = new MultiM(Calc.fMC, n +1, 1);
				MultiM s = new MultiM(Calc.fMC);
				MultiM t0 = new MultiM(Calc.fMC);
				MultiM b = new MultiM(Calc.fMC, n, n);
				MultiM c = new MultiM(Calc.fMC, n, n);
				if (n % 2 != 0)
					zz.E[0, 0] &= MultiFD.fMOne;
				else
					zz.E[0, 0] &= MultiFD.fOne;
				for (m = 0; m < n; m++) 
				{
					if (m == 0) 
					{
						for (i = 0; i < n; i++)
							for (j = 0; j < n; j++)
								c.E[i, j] &= x.E[i, j];
					}
					else 
					{
						for (i = 0; i < n; i++) 
						{
							for (j = 0; j < n; j++) 
							{
								s &= MultiFD.fZero;
								for (k = 0; k < n; k++)
									s &= s + b.E[i, k] * x.E[k, j];
								c.E[i, j] &= s;
							}
						}
					}
					Trace(c, t0);
					Calc.SetToNM(t0 / (m + 1), t0);
					zz.E[m + 1, 0] &= -t0 * zz.E[0, 0];
					if (m < n - 1)
						for (i = 0; i < n; i++)
							for (j = 0; j < n; j++)
								if (j == i)
									b.E[i, j] &= c.E[i, j] - t0;
								else
									b.E[i, j] &= c.E[i, j];
				}
				if (n % 2 != 0)
					zz &= -zz;
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.SetToNM(zz, z);
			}
		} // CharPoly

		public static void CharPolyN(MultiM x, int zI)
		// z = Characteristic polynomial of square matrix x
		{
			if (x.Cs != x.Rs)
				MultiMD.ErrNotSquare();
			else
			{
				MultiM z = new MultiM();
				CharPoly(x, z);
				Calc.SetToN(z, zI);
			}
		} // CharPolyN

		public static void CompM(MultiM p, MultiM z)
		// z = companion matrix of polynomial p
		{
			if (p.Cs != 1)
				MultiMD.ErrNotColumn();
			else
			{
				PolMonic(p);
				int r;
				int n = p.Rs - 1;
				MultiM M = new MultiM(1, n, n);
				for (r = 0; r < n; r++)
					XPCalLst.SetToNM(-p.E[n - r, 0], M.E[r, n - 1]);
				for (r = 1; r < n; r++)
					XPCalLst.SetToNM(MultiFD.fOne, M.E[r, r - 1]);
				Calc.SetToNM(M, z);
			}
		} // CompM

		public static void CompMN(MultiM p, int zI)
		// z = companion matrix of polynomial p
		{
			if (p.Cs != 1)
				MultiMD.ErrNotColumn();
			else
			{
				MultiM z = new MultiM();
				CompM(p, z);
				Calc.SetToN(z, zI);
			}
		} // CompMN

		public static void PolMonic(MultiM z)
		// z = Monic polynomial of z, leading coeff. = 1 or {0}
		{
			if (z.Cs != 1)
				MultiMD.ErrNotColumn();
			else
			{
				PolNorm(z);
				if (z.E[0, 0] == MultiFD.fZero || z.E[0, 0] == MultiFD.fOne)
					return;
				for (int r = 1; r < z.Rs; r++)
					XPCalLst.SetToNM(z.E[r, 0] / z.E[0, 0], z.E[r, 0]);
				XPCalLst.SetToNM(MultiFD.fOne, z.E[0, 0]);
			}
		} // PolMonic

		public static void PolMonicN(MultiM x, int zI)
		// z = Monic polynomial of z, leading coeff. = 1 or {0}
		{
			if (x.Cs != 1)
				MultiMD.ErrNotColumn();
			else
			{
				PolMonic(x);
				Calc.SetToN(x, zI);
			}
		} // PolMonicN

		public static void PolNorm(MultiM z)
		// z = norm of polynomial z, leading coeff. not zero or {0}
		{
			if (z.Cs != 1)
				MultiMD.ErrNotColumn();
			else
			{
				if (z.E[0, 0] != MultiFD.fZero)
					return;
				int nz = 0;
				int r;
				for (r = 0; r < z.Rs - 1; r++)
					if (z.E[r, 0] == MultiFD.fZero)
						nz++;
					else
						break;
				MultiM zz = new MultiM(1, z.Rs - nz, 1);
				for (r = nz; r < z.Rs; r++)
					XPCalLst.SetToNM(z.E[r, 0], zz.E[r - nz, 0]);
				Calc.SetToNM(zz, z);
			}
		} // PolNorm

		public static void PolNormN(MultiM x, int zI)
		// z = norm of polynomial x, leading coeff. not zero or {0}
		{
			if (x.Cs != 1)
				MultiMD.ErrNotColumn();
			else
			{
				PolNorm(x);
				Calc.SetToN(x, zI);
			}
		} // PolNormN

		public static void PolAdd(MultiM u, MultiM v, MultiM z) // z = polynomial u + v
		{
			if (u.Cs != 1 || v.Cs != 1)
				MultiMD.ErrNotColumn();
			else
			{
				int m = u.Rs;
				int n = v.Rs;
				int p = Math.Max(m, n);
				MultiM zz = new MultiM(Calc.fMC, p, 1);
				int i, j, k;
				for (i = m - 1, j = n - 1, k = p - 1; i >= 0 && j >= 0; i--, j--, k--)
					XPCalLst.SetToNM(u.E[i, 0] + v.E[j, 0], zz.E[k, 0]);
				if (m >= n)
					for (i = 0; i < m - n; i++)
						XPCalLst.SetToNM(u.E[i, 0], zz.E[i, 0]);
				else
					for (i = 0; i < n - m; i++)
						XPCalLst.SetToNM(v.E[i, 0], zz.E[i, 0]);
				PolNorm(zz);
				Calc.SetToNM(zz, z);
			}
		} // PolAdd

		public static void PolAddN(MultiM u, MultiM v, int zI) // z = polynomial u + v
		{
			if (u.Cs != 1 || v.Cs != 1)
				MultiMD.ErrNotColumn();
			else
			{
				MultiM z = new MultiM();
				PolAdd(u, v, z);
				Calc.SetToN(z, zI);
			}
		} // PolAddN

		public static void PolMod(MultiM u, MultiM v, MultiM z)
		// if (div)  z = polynomial u/v, "Re" = remainder
		// if (!div) z = "Re" = remainder of u/v
		{
			PolDiv(u, v, false, z);
		} // PolMod

		public static void PolDisc(MultiM p, MultiM z) // z = discriminant of polynomial p
		{
			PolNorm(p);
			if (p.Cs != 1)
				MultiMD.ErrNotColumn();
			else if (p.Rs == 1)
				Calc.SetToNM(MultiMD.mZero, z);
			else if (p.Rs == 2)
				Calc.SetToNM(MultiMD.mOne, z);
			else
			{
				int r;
				int c;
				int n = p.Rs - 1;
				int m = n + n - 1;
				MultiM M = new MultiM(1, m, m);
				for (r = 0; r < n - 1; r++)
					for (c = r; c <= r + n; c++)
						XPCalLst.SetToNM(p.E[c - r, 0], M.E[r, c]);
				for (r = n - 1; r < m; r++)
					for (c = r - n + 1; c <= r; c++)
						XPCalLst.SetToNM((r - c + 1) * p.E[c - r + n - 1, 0], M.E[r, c]);
				Determinant(M, z);
				if (n/2 % 2 == 1)
					MultiMD.ChSM(z);
				Calc.SetToNM(z / p.E[0, 0], z);
			}
		} // PolDisc

		public static void PolDiscN(MultiM p, int zI) // z = discriminant of polynomial p
		{
			if (p.Cs != 1)
				MultiMD.ErrNotColumn();
			else
			{
				MultiM z = new MultiM();
				PolDisc(p, z);
				Calc.SetToN(z, zI);
			}
		} // PolDiscN

		public static void PolDiv(MultiM u, MultiM v, bool div, MultiM z)
		// if (div)  z = polynomial u/v, "Re" = remainder
		// if (!div) z = "Re" = remainder of u/v
		{
			if (u.Cs != 1 || v.Cs != 1)
				MultiMD.ErrNotColumn();
			else
			{

				int reI;
				Common.LFindAdd("Re", out reI);
				PolNorm(u);
				PolNorm(v);
				if (u.Rs < v.Rs)
				{
					Calc.SetToN(u, reI);
					if (div)
						Calc.SetToNM(MultiMD.vZero, z);
					else
						Calc.SetToNM(u, z);
				}
				else
				{
					bool saveExp = MultiFD.expanded;
					int saveFMC = Calc.fMC;
					int saveMMax = MultiID.mMax;
					XPCalLst.Expand();
					int m = u.Rs;
					int n = v.Rs;
					int p = m - n + 1;
					MultiM q = new MultiM(Calc.fMC, p, 1);
					MultiM r = new MultiM(Calc.fMC, m, 1);
					int k, j;
					for (j = 0; j < m; j++) 
						r.E[j, 0] &= u.E[j, 0];
					for (j = 0; j < p; j++) 
						q.E[j, 0] &= MultiFD.fZero;
					for (k = 0; k <= m - n; k++) 
					{
						q.E[k, 0] &= r.E[k, 0] / v.E[0, 0];
						if (MultiID.muErr || MultiID.muAbort)
						{
							MultiFD.expanded = saveExp;
							Calc.fMC = saveFMC;
							MultiID.mMax = saveMMax;
							return;
						}
						for (j = k + 1; j < n + k; j++)
							r.E[j, 0] -= q.E[k, 0] * v.E[j - k, 0];
					}
					for (j = 0; j < m - n + 1; j++)
						r.E[j, 0] &= MultiFD.fZero;
					PolNorm(r);
					MultiFD.expanded = saveExp;
					Calc.fMC = saveFMC;
					MultiID.mMax = saveMMax;
					Calc.SetToN(r, reI);
					if (div)
						Calc.SetToNM(q, z);
					else
						Calc.SetToNM(r, z);
				}
			}
		} // PolDiv

		public static void PolDivN(MultiM u, MultiM v, bool div, int zI)
		// if (div)  z = polynomial u/v, "Re" = remainder
		// if (!div) z = "Re" = remainder of u/v
		{
			if (u.Cs != 1 || v.Cs != 1)
				MultiMD.ErrNotColumn();
			else
			{
				MultiM z = new MultiM();
				PolDiv(u, v, div, z);
				Calc.SetToN(z, zI);
			}
		} // PolDivN

		public static void PolEval(MultiM p, MultiM x, MultiM z) // z = polynomial evaluated at x
		{
			if (p.Cs != 1)
				MultiMD.ErrNotColumn();
			else if (x.Rs != 0)
				MultiMD.ErrNotScalar();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int n = p.Rs;
				MultiM zz = new MultiM(Calc.fMC);
				zz &= p.E[0, 0];
				for (int i = 1; i < n; i++)
					zz &= zz * x + p.E[i, 0];
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.SetToNM(zz, z);
			}
		} // PolEval

		public static void PolEvalN(MultiM p, MultiM x, int zI) // z = polynomial evaluated at x
		{
			if (p.Cs != 1)
				MultiMD.ErrNotColumn();
			else if (x.Rs != 0)
				MultiMD.ErrNotScalar();
			else
			{
				MultiM z = new MultiM();
				PolEval(p, x, z);
				Calc.SetToN(z, zI);
			}
		} // PolEvalN

		public static void PolMul(MultiM u, MultiM v, MultiM z) // z = polynomial u*v
		{
			if ((u.Cs == 0 && v.Cs == 1) || (u.Cs == 1 && v.Cs == 0))
			{
				MultiM zz = new MultiM(Calc.fMC, Math.Max(u.Rs, v.Rs), 1);
				MultiMD.MulM(u, v, zz);	// polynomial * a scalar
				Calc.SetToNM(zz, z);
				return;
			}
			if (u.Cs != 1 || v.Cs != 1)
				MultiMD.ErrNotColumn();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int m = u.Rs;
				int n = v.Rs;
				MultiM zz = new MultiM(Calc.fMC, n + m - 1, 1);
				for (int i = 0; i < m; i++)
					for (int j = 0; j < n; j++)
						zz.E[i + j, 0] += u.E[i, 0] * v.E[j, 0];
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.SetToNM(zz, z);
			}
		} // PolMul

		public static void PolMulN(MultiM u, MultiM v, int zI)	// z = polynomial u*v
		{
			MultiM z = new MultiM();
			PolMul(u, v, z);
			Calc.SetToN(z, zI);
		} // PolMulN

		public static void PolRecip(MultiM p, MultiM z) // z = reciprocal polynomial of p
		// p is a polynomial, z = x^deg(p) * p(1/x), the coefficients are reversed
		{
			if (p.Cs != 1)
				MultiMD.ErrNotColumn();
			else
			{
				int n = p.Rs - 1;
				MultiM zz = new MultiM(1, p.Rs, 1);
				for (int r = 0; r <= n; r++)
					XPCalLst.SetToNM(p.E[r, 0], zz.E[n - r, 0]);
				Calc.SetToNM(zz, z);
			}
		} // PolRecip

		public static void PolRecipN(MultiM p, int zI) // z = reciprocal polynomial of p
		// p is a polynomial, z = x^deg(p) * p(1/x), the coefficients are reversed
		{
			if (p.Cs != 1)
				MultiMD.ErrNotColumn();
			else
			{
				MultiM z = new MultiM();
				PolRecip(p, z);
				Calc.SetToN(z, zI);
			}
		} // PolRecipN

		public static void PolSturm(MultiM p, MultiM z) // z = number of real roots in p
		// p is a polynomial with real coeffs. should be square free
		{
			if (p.Cs != 1)
				MultiMD.ErrNotColumn();
			else
			{
				PolMonic(p);
				MultiM a = new MultiM(Calc.fMC);
				a &= MultiMD.mOne;
				for (int r = 1; r < p.Rs; r++)
					a &= a + Abs(p.E[r, 0]);
				Calc.NormNM(a);
				PolSturmI(p, -a, a, z);
			}
		} // PolSturm

		public static void PolSturmN(MultiM p, int zI) // z = number of real roots in p
		// p is a polynomial with real coeffs. should be square free
		{
			if (p.Cs != 1)
				MultiMD.ErrNotColumn();
			else
			{
				MultiM z = new MultiM();
				PolSturm(p, z);
				Calc.SetToN(z, zI);
			}
		} // PolSturmN

		public static void PolSturmI(MultiM p, MultiM a, MultiM b, MultiM z) // z = number of real roots in p
		// in interval (a, b). roots at a or b may or may not be counted
		// p is a polynomial with real coeffs. should be square free
		{
			if (p.Cs != 1)
				MultiMD.ErrNotColumn();
			else if (a.Rs != 0 || b.Rs != 0)
				MultiMD.ErrNotScalar();
			else
			{
				PolMonic(p);
				if (p.Rs == 1)
					z &= MultiMD.mZero;
				else
				{
					int r;
					int d = 0;
					MultiM f0 = new MultiM(1, p.Rs, 1);
					MultiM f1 = new MultiM(1, p.Rs - 1, 1);
					MultiM f2 = new MultiM(1, p.Rs - 2, 1);
					MultiM v0 = new MultiM(2);
					MultiM v1 = new MultiM(2);
					Calc.SetToNM(p, f0);
					for (r = 0; r < p.Rs - 1; r++)
						XPCalLst.SetToNM((p.Rs - 1 - r) * f0.E[r, 0], f1.E[r, 0]);
					PolEval(f0, a, v0);
					if(v0 == MultiFD.fZero)
						MultiID.WriteLn("PolSturmI: PolEval(p, a) = 0.0");
					PolEval(f1, a, v1);
					if (v0.S != v1.S)
						d++;
					while (f1.Rs > 1)
					{
						PolMod(-f0, f1, f2);
						Calc.SetToNM(f1, f0);
						Calc.SetToNM(v1, v0);
						Calc.SetToNM(f2, f1);
						PolEval(f1, a, v1);
						if (v0.S != v1.S)
							d++;			
					}
					if (f1.E[0, 0] == MultiFD.fZero)
					{
						MultiID.MuWriteErr("PolSturm failed: Polynomial is not square free");
						return;
					}
					Calc.SetToNM(p, f0);
					f1 = new MultiM(1, p.Rs - 1, 1);
					for (r = 0; r < p.Rs - 1; r++)
						 XPCalLst.SetToNM((p.Rs - 1 - r) * f0.E[r, 0], f1.E[r, 0]);
					PolEval(f0, b, v0);
					if(v0 == MultiFD.fZero)
						MultiID.WriteLn("PolSturmI: PolEval(p, b) = 0.0");
					PolEval(f1, b, v1);
					if (v0.S != v1.S)
						d--;
					while (f1.Rs > 1)
					{
						PolMod(-f0, f1, f2);
						Calc.SetToNM(f1, f0);
						Calc.SetToNM(v1, v0);
						Calc.SetToNM(f2, f1);
						PolEval(f1, b, v1);
						if (v0.S != v1.S)
							d--;			
					}
					z &= Math.Abs(d);
				}
				Calc.NormNM(z);
			}
		} // PolSturmI

		public static void PolSturmIN(MultiM p, MultiM a, MultiM b, int zI) // z = number of real roots in p
		// in interval (a, b). roots at a or b may or may not be counted
		// p is a polynomial with real coeffs. should be square free
		{
			if (p.Cs != 1)
				MultiMD.ErrNotColumn();
			else if (a.Rs != 0 || b.Rs != 0)
				MultiMD.ErrNotScalar();
			else
			{
				MultiM z = new MultiM();
				PolSturmI(p, a, b, z);
				Calc.SetToN(z, zI);
			}
		} // PolSturmIN

		public static void SetAll(MultiM m, MultiM x, int zI) // z = m, z[r, c] = x for all z[r, c]
		{
			if (x.Rs != 0)
				MultiMD.ErrNotScalar();
			else if (m.Rs == 0)
				MultiMD.ErrElementOfScalar();
			else
			{
				for (int r = 0; r < m.Rs; r++)
					for (int c = 0; c < m.Cs; c++)
					{
						XPCalLst.SetToNM((MultiF)x, m.E[r, c]);
					}
				Calc.SetToN(m, zI);
			}
		} // SetAll

		public static void SetElement(MultiM m, MultiM rF, MultiM cF, MultiM x, int zI)	// z = m, z[r, c] = x
		{
			if (rF.Rs != 0 || cF.Rs != 0 || x.Rs != 0)
				MultiMD.ErrNotScalar();
			else if (m.Rs == 0)
				MultiMD.ErrElementOfScalar();
			else
			{
				int r;
				int c;
				MultiFD.Get1F(out r, rF);
				MultiFD.Get1F(out c, cF);
				r -= uFirst;
				c -= uFirst;
				Calc.SetToN(m, zI);
				if (r >= 0 && c >= 0 && r < Common.nList[zI].Rs && c < Common.nList[zI].Cs)
					XPCalLst.SetToNM(x, Common.nList[zI].E[r, c]);
				else
					MultiMD.ErrOutside();
			}
		} // SetElement

		public static void GetElement(MultiM x, MultiM rF, MultiM cF, int zI) // z = x[r, c]
		{
			if (rF.Rs != 0 || cF.Rs != 0)
				MultiMD.ErrNotScalar();
			else
			{
				int r;
				int c;
				MultiFD.Get1F(out r, rF);
				MultiFD.Get1F(out c, cF);
				GetElement(x, r, c, zI);
			}
		} // GetElement

		public static void GetElement(MultiM x, int ri, int ci, int zI)	// z = x[r, c]
		{
			if (x.Rs == 0)
				MultiMD.ErrNotMatrix();
			else
			{
				int r = ri - uFirst;
				int c = ci - uFirst;
				if (r >= 0 && c >= 0 && r < x.Rs && c < x.Cs)
				{
					MultiM z = new MultiM();
					Calc.SetToNM(x.E[r, c], z);
					Calc.SetToN(z, zI);
				}
				else
					MultiMD.ErrOutside();
			}
		} // GetElement

		public static MultiM Transpose(MultiM x)	// return x'
		{
			if (x.Rs == 0)
			{
				MultiMD.ErrNotMatrix();
				return MultiMD.mZero;
			}
			else
			{
				MultiM z = new MultiM(0, x.Cs, x.Rs);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
						XPCalLst.SetToNM(x.E[r, c], z.E[c, r]);
				return z;
			}
		} // Transpose

		public static void TransposeN(int zI)	// z = z'
		{
				Calc.SetToN(Transpose(Common.nList[zI]), zI);
		} // TransposeN

		public static MultiM Adjoint(MultiM x)	// return Adjoint matrix of square matrix x
		{
			if (x.Rs == 0 || x.Rs != x.Cs)
			{
				MultiMD.ErrNotSquare();
				return MultiMD.mZero;
			}
			else if (x.Rs == 1)
			{
				MultiM z = new MultiM(1, 1, 1); // 1 by 1 identity matrix
				z.E[0, 0] = MultiFD.fOne;
				return z;
			}
			else
			{
				int n = x.Rs;
				int nm = n - 1;
				int mI;
				Common.InitNAdd(1, "Adjoint-mI", out mI);
				MultiM z = new MultiM(0, n, n);
				for (int h = 0; h < n; h++)
					for (int j = 0; j < n; j++)
					{
						MatrixN(nm, nm, mI);
						for (int r = 0; r < nm; r++)
							for (int c = 0; c < nm; c++)
							{
								int r2 = r;
								int c2 = c;
								if (r2 >= h) r2++;
								if (c2 >= j) c2++;
								XPCalLst.SetToNM(x.E[r2, c2], Common.nList[mI].E[r, c]);
							}
						DeterminantN(Common.nList[mI], mI);
						XPCalLst.SetToNM(Common.nList[mI], z.E[j, h]);
						int odd;
						Math.DivRem(h + j, 2, out odd);
						if (odd == 1)
							MultiFD.ChSF(z.E[j, h]);
					}
				return z;
			}
		} // Adjoint

		public static void AdjointN(int zI)	// z = Adjoint matrix of square matrix z
		{
			Calc.SetToN(Adjoint(Common.nList[zI]), zI);
		} // AdjointN

		public static MultiM Balance(MultiM x)	// return a balanced matrix with identical eigenvalues as x
		// Given a matrix x[1..n][1..n], this routine replaces it by a balanced matrix with identical
		// eigenvalues. A symmetric matrix is already balanced and is unaffected by this procedure. The
		// parameter radix should be the machine's floating-point radix.
		{
			if (x.Rs == 0 || x.Rs != x.Cs)
			{
				MultiMD.ErrNotSquare();
				return MultiMD.mZero;
			}
			else if (x.Rs == 1)
			{
				MultiM z = new MultiM(1, 1, 1); // 1 by 1 matrix
				Calc.SetToNM(x, z);
				return z;
			}
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int n = x.Rs;
				int i, j, last = 0;
				MultiM z = new MultiM(Calc.fMC, n, n);
				MultiF s = new MultiF(Calc.fMC);
				MultiF r = new MultiF(Calc.fMC);
				MultiF g = new MultiF(Calc.fMC);
				MultiF f = new MultiF(Calc.fMC);
				MultiF c = new MultiF(Calc.fMC);
				MultiF radix = new MultiF(1);
				MultiF sqrdx = new MultiF(1);
				MultiFD.SetTo1F(10, radix);
				sqrdx &= radix * radix;
				MultiF cP95 = new MultiF(1);
				MultiFD.ValueF10("0.95", out i, cP95);
				z &= x;
				while (last == 0) 
				{
					last = 1;
					for (i = 0; i < n; i++)	// Calculate row and column norms.
					{
						r &= c &= MultiFD.fZero;
						for (j = 0; j < n; j++)
							if (j != i)
							{
								c += Abs(z.E[j, i]);
								r += Abs(z.E[i, j]);
							}
						if (c != MultiFD.fZero && r != MultiFD.fZero)	// if both are nonzero
						{
							g &= r / radix;
							f &= MultiFD.fOne;
							s &= c + r;
							while (c < g)	// find the integer power of the machine radix that
							{				// comes closest to balancing the matrix.
								f *= radix;
								c *= sqrdx;
							}
							g &= r * radix;
							while (c > g) 
							{
								f /= radix;
								c /= sqrdx;
							}
							if ((c + r) / f < cP95 * s) 
							{
								last = 0;
								g &= MultiFD.fOne / f;
								for (j = 0; j < n; j++) z.E[i, j] *= g;	// Apply similarity transformation.
								for (j = 0; j < n; j++) z.E[j, i] *= f;
							}
						}
					}
				}
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormNM(z);
				return z;
			}
		} // Balance

		public static void BalanceN(int zI)	// replace z by a balanced matrix with identical eigenvalues
		{
			Calc.SetToN(Balance(Common.nList[zI]), zI);
		} // BalanceN

		public static void SwapC(ref MultiF x, ref MultiF y)	// interchange x and y (pointers only)
		{
			MultiF t;
			t = x;
			x = y;
			y = t;
		} // SwapC

		public static MultiM Hess(MultiM a)	// return upper Hessenberg matrix with identical eigenvalues as a
		// Reduction to Hessenberg form by the elimination method
		// Recommended, but not required, is that this routine be preceded by Balance
		{
			if (a.Rs == 0 || a.Rs != a.Cs)
			{
				MultiMD.ErrNotSquare();
				return MultiMD.mZero;
			}
			else if (a.Rs <= 2)
			{
				MultiM z = new MultiM(); // 1 by 1 or 2 by 2 matrix
				Calc.SetToNM(a, z);
				return z;
			}
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int n = a.Rs;
				int m, j, i;
				MultiM z = new MultiM(Calc.fMC, n, n);
				MultiF y = new MultiF(Calc.fMC);
				MultiF x = new MultiF(Calc.fMC);
				z &= a;
				for (m = 1; m < n - 1; m++)	// m is called r + 1 in the text.
				{
					x &= MultiFD.fZero;
					i = m;
					for (j = m; j < n; j++)	// Find the pivot.
					{
						if (Abs(z.E[j, m - 1]) > Abs(x))
						{
							x &= z.E[j, m - 1];
							i = j;
						}
					}
					if (i != m)	// Interchange rows and columns.
					{
						for (j = m - 1; j < n; j++) SwapC(ref z.E[i, j], ref z.E[m, j]);
						for (j = 0; j < n; j++) SwapC(ref z.E[j, i], ref z.E[j, m]);
					}
					if (x != MultiFD.fZero)	// Carry out the elimination.
					{
						for (i = m + 1; i < n; i++) 
						{
							if ((y &= z.E[i, m - 1]) != MultiFD.fZero) 
							{
								y /= x;
								z.E[i, m - 1] &= y;
								for (j = m; j < n; j++)
									z.E[i, j] -= y * z.E[m, j];
								for (j = 0; j < n; j++)
									z.E[j, m] += y * z.E[j, i];
							}
						}
					}
				}
				for (j = 0; j < n - 2; j++)
					for (i = j + 2; i < n; i++)
						z.E[i, j] &= MultiFD.fZero;
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormNM(z);
				return z;
			}
		} // Hess

		public static void HessN(int zI)	// replace z by a balanced matrix with identical eigenvalues
		{
			Calc.SetToN(Hess(Common.nList[zI]), zI);
		} // HessN

		public static void Deriv(MultiM p, MultiM z)	// z = derivative of polynomial p
		{
			if (p.Cs != 1)
				MultiMD.ErrNotColumn();
			else
			{
				int n = p.Rs;
				MultiM zz = new MultiM(1, p.Rs - 1, 1);
				for (int r = 0; r < n - 1; r++)
					XPCalLst.SetToNM((p.Rs - 1 - r) * p.E[r, 0], zz.E[r, 0]);
				Calc.SetToNM(zz, z);
			}
		} // Deriv

		public static void DerivN(MultiM p, int zI)	// z = derivative of polynomial p
		{
			if (p.Cs != 1)
				MultiMD.ErrNotColumn();
			else
			{
				MultiM z = new MultiM();
				Deriv(p, z);
				Calc.SetToN(z, zI);
			}
		} // DerivN

		public static void Integ(MultiM p, MultiM z)	// z = formal integral of polynomial p
		{
			if (p.Cs != 1)
				MultiMD.ErrNotColumn();
			else
			{
				int n = p.Rs;
				MultiM zz = new MultiM(1, p.Rs + 1, 1);
				for (int r = 0; r < n; r++)
					XPCalLst.SetToNM(p.E[r, 0] / (n - r), zz.E[r, 0]);
				Calc.SetToNM(zz, z);
			}
		} // Integ

		public static void IntegN(MultiM p, int zI)	// z = formal integral of polynomial p
		{
			if (p.Cs != 1)
				MultiMD.ErrNotColumn();
			else
			{
				MultiM z = new MultiM();
				Integ(p, z);
				Calc.SetToN(z, zI);
			}
		} // IntegN

		public static void Determinant(MultiM x, MultiM z)	// z = |x|
		{
			int rs = x.Rs;
			int cs = x.Cs;
			if (rs == 0 || rs != cs)
				MultiMD.ErrNotSquare();
			else if (rs == 1)
			{
				Calc.SetToNM(x.E[0, 0], z);
			}
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiF d = new MultiF(Calc.fMC);
				if (rs == 2)
				{
					d &= x.E[0, 0] * x.E[1, 1] - x.E[0, 1] * x.E[1, 0];
				}
				else
				{
					int n = rs;
					MultiM a = new MultiM(Calc.fMC, n, n);
					int[] index;
					index = new int[n];
					a &= x;
					bool sing;
					MultiMD.LUdecomp(a, index, d, out sing); // decompose the matrix
					if (sing)
						d &= MultiFD.fZero;
					else
						for (int i = 0; i < n; i++)
							d *= a.E[i, i];
				}
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.SetToNM(d, z);
			}
		} // Determinant

		public static void DeterminantN(MultiM x, int zI)	// z = |x|
		{
			int rs = x.Rs;
			int cs = x.Cs;
			if (rs == 0 || rs != cs)
				MultiMD.ErrNotSquare();
			else
			{
				MultiM z = new MultiM();
				Determinant(x, z);
				Calc.SetToN(z, zI);
			}
		} // DeterminantN

		public static void NotN(MultiM x, int zI)	// z = !x
		{
			if (x.Rs == 0)
				XPCalLst.NotN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "NotN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.NotN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						XPCalLst.SetToNM((MultiF)Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // NotN

		public static void RNotN(int zI) // z = !z
		{
			NotN(Common.nList[zI], zI);
		}

		public static void Primorial(int zI) // change sign, z = - z
		{
			if (Common.nList[zI].Rs == 0)
			{
				XPCalLst.RIntSIN(zI);
				XICommon.Primorial(zI);
				Calc.NormN(zI);
			}
			else
			{
				int tI;
				Common.InitNAdd(1, "Primorial-tI", out tI);
				for (int r = 0; r < Common.nList[zI].Rs; r++)
					for (int c = 0; c < Common.nList[zI].Cs; c++)
					{
						Calc.SetToN(Common.nList[zI].E[r, c], tI);
						XPCalLst.RIntSIN(tI);
						XICommon.Primorial(tI);
						XPCalLst.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
						XPCalLst.NormNM(Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // Primorial

		public static void BinoS(int xI, int zI)
		// z = Binomial coefficient (x, z) slow method
		// a, b are not modified unless they share memory with z
		// will write to standard output if diagOn is true		
		{
			int xrs = Common.nList[xI].Rs;
			int zrs = Common.nList[zI].Rs;
			int xcs = Common.nList[xI].Cs;
			int zcs = Common.nList[zI].Cs;
			if (xrs == 0 && zrs == 0)
			{
				XPCalLst.RIntSIN(zI);
				XPCalLst.RIntSIN(xI);
				XICommon.BinoS(Common.nList[zI], Common.nList[xI], Common.nList[zI]);
				Calc.NormN(zI);
			}
			else if (xrs != zrs || xcs != zcs)
				MultiMD.ErrIncompatible();
			else
			{
				int tI, t2I;
				Common.InitNAdd(1, "BinoS-tI", out tI);
				Common.InitNAdd(1, "BinoS-t2I", out t2I);
				for (int r = 0; r < zrs; r++)
					for (int c = 0; c < zcs; c++)
					{
						Calc.SetToN(Common.nList[zI].E[r, c], tI);
						Calc.SetToN(Common.nList[xI].E[r, c], t2I);
						XPCalLst.RIntSIN(tI);
						XPCalLst.RIntSIN(t2I);
						Common.nList[zI].E[r, c].C = 0;
						XICommon.BinoS(Common.nList[tI], Common.nList[t2I], Common.nList[zI].E[r, c]);
						XPCalLst.NormNM(Common.nList[zI].E[r, c]);
					}
				Common.LDelete(t2I);
				Common.LDelete(tI);
			}
		} // BinoS

		public static bool swapped;		// true if SortC or SortR swapped elements
		public static int uFirst;		// the user index of the first element in a matrix
	}
}
